from __future__ import annotations

import asyncio

import attrs
from datetime import datetime
import logging
import nibabel as nib
from nibabel.affines import apply_affine
import numpy as np
import os
import pathlib
import pyvista as pv
import pyvistaqt as pvqt
import qtawesome as qta
from qtpy import QtWidgets, QtGui, QtCore
import shutil
import typing as tp

from . import MainViewPanel
from NaviNIBS.Navigator.GUI.Widgets.MRIViews import MRISliceView, MRI3DView
from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.GUI.Icons import getIcon
from NaviNIBS.util.GUI.QFileSelectWidget import QFileSelectWidget
from NaviNIBS.util.GUI.QMouseWheelAdjustmentGuard import preventAnnoyingScrollBehaviour
from NaviNIBS.util.Signaler import Signal


logger = logging.getLogger(__name__)


@attrs.define()
class MRIPanel(MainViewPanel):
    _key: str = 'Set MRI'
    _icon: QtGui.QIcon = attrs.field(init=False, factory=lambda: getIcon('mdi6.image'))
    _filepathWdgt: QFileSelectWidget = attrs.field(init=False)
    _views: tp.Dict[str, tp.Union[MRISliceView, MRI3DView]] = attrs.field(init=False, factory=dict)

    _climSpinboxWidgets: dict[str, QtWidgets.QDoubleSpinBox] = attrs.field(init=False, factory=dict)
    _climCheckboxWidgets: dict[str, QtWidgets.QCheckBox] = attrs.field(init=False, factory=dict)

    finishedAsyncInit: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def canBeEnabled(self) -> tuple[bool, str | None]:
        if self.session is None:
            return False, 'No session set'
        return True, None

    def _finishInitialization(self):
        # don't initialize computationally-demanding views until panel is activated (viewed)

        super()._finishInitialization()

        self._wdgt.setLayout(QtWidgets.QVBoxLayout())

        wdgt = QFileSelectWidget(browseMode='getOpenFilename',
                                 extFilters='Nifti (*.nii; *.nii.gz)')
        # TODO: set supported file formats to (.nii | .nii.gz) only
        self._wdgt.layout().addWidget(wdgt)
        self._filepathWdgt = wdgt

        containerWdgt = QtWidgets.QWidget()
        containerWdgt.setSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        containerLayout = QtWidgets.QGridLayout()
        containerWdgt.setLayout(containerLayout)
        self._wdgt.layout().addWidget(containerWdgt)

        for iDim, dim in enumerate(('2D', '3D')):
            dimOuterContainer = QtWidgets.QWidget()
            dimOuterContainer.setLayout(QtWidgets.QVBoxLayout())
            containerWdgt.layout().addWidget(dimOuterContainer, iDim, 2)
            dimContainer = QtWidgets.QGroupBox(f'{dim} colorbar limits')
            dimContainer.setLayout(QtWidgets.QGridLayout())
            dimContainer.setContentsMargins(0, 0, 0, 0)
            dimOuterContainer.layout().addWidget(dimContainer)
            dimOuterContainer.layout().addStretch()

            for iRow, minOrMax in enumerate(('Max', 'Min')):
                wdgt = QtWidgets.QLabel(f'{minOrMax}:')
                dimContainer.layout().addWidget(wdgt, iRow, 0)

                wdgt = QtWidgets.QDoubleSpinBox()
                preventAnnoyingScrollBehaviour(wdgt)
                wdgt.editingFinished.connect(
                    lambda *args, dim=dim, minOrMax=minOrMax:
                    self._onClimSpinboxChanged(dim, minOrMax))
                wdgt.setDecimals(1)
                wdgt.setMinimum(0)
                wdgt.setMaximum(float('inf'))
                wdgt.setMinimumWidth(100)
                dimContainer.layout().addWidget(wdgt, iRow, 1)
                self._climSpinboxWidgets[f'clim{dim}{minOrMax}'] = wdgt

                wdgt = QtWidgets.QLabel('Auto:')
                dimContainer.layout().addWidget(wdgt, iRow, 3)

                wdgt = QtWidgets.QCheckBox()
                wdgt.stateChanged.connect(
                    lambda *args, dim=dim, minOrMax=minOrMax:
                    self._onClimCheckboxChanged(dim, minOrMax))
                dimContainer.layout().addWidget(wdgt, iRow, 4)
                self._climCheckboxWidgets[f'clim{dim}{minOrMax}'] = wdgt

            self._updateClimWidgetsFromModel(dim)

        for iRow, iCol, dim in ((0, 1, 'x'), (0, 0, 'y'), (1, 0, 'z'), (1, 1, '3D')):
            if dim in ('x', 'y', 'z'):
                self._views[dim] = MRISliceView(normal=dim, doShowScalarBar=True)
            elif dim == '3D':
                self._views[dim] = MRI3DView(label=dim, doShowScalarBar=True)
            else:
                raise NotImplementedError()

            self._views[dim].sigSliceOriginChanged.connect(lambda key=dim: self._onSliceOriginChanged(sourceKey=key))

            containerLayout.addWidget(self._views[dim].wdgt, iRow, iCol)

        if self.session is not None:
            self._onPanelInitializedAndSessionSet()

    def _onSliceOriginChanged(self, sourceKey: str):
        for key, view in self._views.items():
            if key == sourceKey:
                continue
            view.sliceOrigin = self._views[sourceKey].sliceOrigin

    def _onSessionSet(self):
        super()._onSessionSet()

        if self._hasInitialized:
            self._onPanelInitializedAndSessionSet()

    def _onPanelInitializedAndSessionSet(self):
        self._updateFilepath()
        self._updateRelativeToPath()
        self._filepathWdgt.sigFilepathChanged.connect(self._onBrowsedNewFilepath)
        self.session.sigInfoChanged.connect(self._onSessionInfoChanged)
        self.session.MRI.sigFilepathChanged.connect(self._updateFilepath)
        self.session.MRI.sigClimChanged.connect(self._updateClimWidgetsFromModel)

        for key, view in self._views.items():
            view.session = self.session

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._finishInitialization_async))

    async def _finishInitialization_async(self):
        for view in self._views.values():
            await view.finishedAsyncInit.wait()

        self.finishedAsyncInit.set()

    def _updateFilepath(self):
        self._filepathWdgt.filepath = self.session.MRI.filepath

    def _onSessionInfoChanged(self, whatChanged: tp.Optional[list[str]] = None):
        if whatChanged is None or 'filepath' in whatChanged:
            self._updateRelativeToPath()

    def _updateRelativeToPath(self):
        self._filepathWdgt.showRelativeTo = os.path.dirname(self.session.filepath)
        self._filepathWdgt.showRelativePrefix = '<session>'

    def _onBrowsedNewFilepath(self, newFilepath: str):
        self.session.MRI.filepath = newFilepath

    def _onClimSpinboxChanged(self, dim: str, minOrMax: str):
        logger.debug(f'Clim spinbox changed for {dim}{minOrMax}')
        valKey = f'clim{dim}{minOrMax}'
        isChecked = self._climCheckboxWidgets[valKey].isChecked()
        if not isChecked:
            setattr(self.session.MRI, valKey, self._climSpinboxWidgets[valKey].value())

    def _onClimCheckboxChanged(self, dim: str, minOrMax: str):
        logger.debug(f'Clim checkbox changed for {dim}{minOrMax}')
        valKey = f'clim{dim}{minOrMax}'
        autovalKey = f'autoClim{dim}{minOrMax}'
        isChecked = self._climCheckboxWidgets[valKey].isChecked()
        self._climSpinboxWidgets[valKey].setEnabled(not isChecked)

        if isChecked:
            val = getattr(self.session.MRI, autovalKey)
            setattr(self.session.MRI, valKey, None)  # clear any previous manually set value
        else:
            val = getattr(self.session.MRI, valKey)
            if val is None:
                val = getattr(self.session.MRI, autovalKey)

        if val is None:
            # can't set value, assume we are initializing and will update again later
            return

        self._climSpinboxWidgets[valKey].setValue(val)

    def _updateClimWidgetsFromModel(self, dim: str):
        logger.debug(f'Updating clim widgets from model for {dim}')
        for minOrMax in ('Min', 'Max'):
            valKey = f'clim{dim}{minOrMax}'
            autovalKey = f'autoClim{dim}{minOrMax}'

            if self.session is None or self.session.MRI.filepath is None:
                self._climSpinboxWidgets[valKey].setEnabled(False)
                self._climCheckboxWidgets[valKey].setEnabled(False)
            else:
                self._climCheckboxWidgets[valKey].setEnabled(True)

                isChecked = getattr(self.session.MRI, valKey) is None
                self._climCheckboxWidgets[valKey].setChecked(isChecked)
                self._climSpinboxWidgets[valKey].setEnabled(not isChecked)

                if isChecked:
                    val = getattr(self.session.MRI, autovalKey)
                else:
                    val = getattr(self.session.MRI, valKey)
                    if val is None:
                        val = getattr(self.session.MRI, autovalKey)

                self._climSpinboxWidgets[valKey].setValue(val)




import attrs
import typing as tp

from qtpy import QtWidgets, QtCore, QtGui

from NaviNIBS.Navigator.GUI.CollectionModels import \
    CollectionTableModel, CollectionTableModelBase,\
    FilteredCollectionModel, \
    K, logger
from NaviNIBS.Navigator.Model.Targets import Targets, Target


@attrs.define(slots=False, kw_only=True)
class FullTargetsTableModel(CollectionTableModel[str, Targets, Target]):
    _collection: Targets = attrs.field(init=False, repr=False)
    _isSelectedAttr: str = 'isSelected'

    def __attrs_post_init__(self):
        self._collection = self._session.targets

        self._boolColumns.extend(['isVisible', 'isHistorical'])
        self._editableColumns.extend(['isVisible', 'key'])

        self._columns = [
            'key',
            'isVisible',
            'isHistorical',
            # TODO: add color
            # TODO: add angle, depthOffset, target coord, entry coord as optional but hidden by default
        ]
        self._attrColumns = self._columns.copy()
        self._columnLabels = dict(
            key='Key',
            isVisible='Visibility',
            isHistorical='Historical',
        )

        self._editableColumnValidators = dict(
            key=lambda _, oldKey, newKey: len(newKey) > 0 and (oldKey == newKey or newKey not in self._session.targets),  # don't allow setting one target to key of another target
        )

        self._collection.sigItemsAboutToChange.connect(self._onCollectionAboutToChange)
        self._collection.sigItemsChanged.connect(self._onCollectionChanged)

        super().__attrs_post_init__()


@attrs.define(slots=False, kw_only=True)
class TargetsTableModel(FilteredCollectionModel[str, Targets, Target]):
    """
    Non-historical targets only
    """
    _proxiedModel: FullTargetsTableModel = attrs.field(init=False)

    def __attrs_post_init__(self):
        self._proxiedModel = FullTargetsTableModel(session=self._session)

        FilteredCollectionModel.__attrs_post_init__(self)

    def filterAcceptsRow(self, sourceRow: int, sourceParent: QtCore.QModelIndex) -> bool:
        target = self._proxiedModel.getCollectionItemFromIndex(sourceRow)
        return not target.isHistorical

    def filterAcceptsColumn(self, source_column: int, source_parent: QtCore.QModelIndex):
        columnKey = self._proxiedModel.columns[source_column]
        return columnKey not in ('isHistorical',)





import asyncio
import json
import logging

import attrs
import numpy as np
import pylsl as lsl
import socket
import time
import threading
import typing as tp


from NaviNIBS.util.ZMQConnector import ZMQConnectorServer, ZMQConnectorClient, getNewPort


logger = logging.getLogger(__name__)


lsl_fmt2string = ['undefined', 'float32', 'double64', 'string', 'int32', 'int16','int8', 'int64']


def getDTypeFromStreamInfo(streamInfo: lsl.StreamInfo) -> np.dtype:
    dataType = lsl_fmt2string[streamInfo.channel_format()]
    if dataType == 'string':
        dtype = 'object'
    elif dataType == 'double64':
        dtype = 'float64'
    else:
        dtype = dataType
    return dtype


def getKeyForStreamInfo(streamInfo: lsl.StreamInfo) -> str:
    # construct what should be a unique key as streamName@hostname
    # (Note that this does not use source_id since that is not reliably known in advance and so wouldn't work for
    #   pre-specifying desired streams. This is in important distinction from the keys used by LSLStreamResolver)
    return '%s@%s' % (streamInfo.name(), streamInfo.hostname())
    # TODO: add support wherever this is used for "localhost" to appear equivalent to actual hostname


def getEquivalentKeysForStreamInfo(streamInfo: lsl.StreamInfo) -> tp.List[str]:
    streamKey = getKeyForStreamInfo(streamInfo)
    equivalentStreamKeys = [streamKey]
    if streamInfo.hostname() == socket.gethostname():
        # allow a local stream by to be specified by name only, or by name@localhost
        equivalentStreamKeys = [streamInfo.name(), '%s@%s' % (streamInfo.name(), 'localhost')] + equivalentStreamKeys
    return equivalentStreamKeys


def inferStreamKey(equivStreamKeys: tp.List[str], preferredStreamKeys: tp.Iterable[str]) -> str:
    if any(equivStreamKey in preferredStreamKeys for equivStreamKey in equivStreamKeys):
        matchingKeys = [equivStreamKey for equivStreamKey in equivStreamKeys if
                        equivStreamKey in preferredStreamKeys]
        if len(matchingKeys) > 1:
            # multiple matching keys
            raise KeyError()
        return matchingKeys[0]
    else:
        return equivStreamKeys[0]  # if no matches, prefer first equivalent key




from __future__ import annotations

import asyncio
from contextlib import contextmanager
import logging
import multiprocessing as mp
import time
import typing as tp

import attrs
import numpy as np
import pyvista as pv
import vtkmodules.vtkCommonTransforms
import zmq
from qtpy import QtWidgets, QtGui, QtCore
from zmq import asyncio as azmq

from NaviNIBS.util import exceptionToStr, makeStrUnique
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.pyvista.RemotePlotting import ActorRef
from NaviNIBS.util.pyvista.RemotePlotting.RemotePlotter import RemotePlotterApp

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@attrs.define
class RemoteActorProxy:
    _actorID: str
    _plotter: RemotePlotterProxyBase

    _mapper: RemoteMapper | None = attrs.field(init=False, default=None)
    _visibility: bool | None = attrs.field(init=False, default=None)

    @property
    def actorID(self):
        return self._actorID

    @property
    def plotter(self):
        return self._plotter

    @property
    def mapper(self):
        return self.GetMapper()

    def SetUserTransform(self, transform: vtkmodules.vtkCommonTransforms.vtkTransform):

        # convert to ndarray since vtkTransform is not pickleable
        transform_ndarray = pv.array_from_vtkmatrix(transform.GetMatrix())

        return self._plotter.setActorUserTransform(self, transform_ndarray)

    def GetVisibility(self) -> bool:
        if self._visibility is None:
            with self._plotter.disallowNonblockingCalls():
                self._visibility = self._plotter._remoteActorCall(self, 'GetVisibility')
        return self._visibility  # by using cached version, we assume we will be the only one to ever change visibility

    def GetMapper(self):
        if self._mapper is None:
            self._mapper = RemoteMapper(self)
        return self._mapper

    def SetVisibility(self, visibility: bool):
        self._visibility = visibility
        return self._plotter._remoteActorCall(self, 'SetVisibility', visibility)

    def VisibilityOn(self):
        self._visibility = True
        return self._plotter._remoteActorCall(self, 'VisibilityOn')

    def VisibilityOff(self):
        self._visibility = False
        return self._plotter._remoteActorCall(self, 'VisibilityOff')

    def SetUseBounds(self, useBounds: bool):
        return self._plotter._remoteActorCall(self, 'SetUseBounds', useBounds)


@attrs.define
class RemoteCameraProxy:
    _plotter: RemotePlotterProxyBase

    @property
    def position(self):
        return self._plotter._remoteCameraGet('position')

    @position.setter
    def position(self, value):
        self._plotter._remoteCameraSet('position', value)

    @property
    def focal_point(self):
        return self._plotter._remoteCameraGet('focal_point')

    @focal_point.setter
    def focal_point(self, value):
        self._plotter._remoteCameraSet('focal_point', value)

    @property
    def up(self):
        return self._plotter._remoteCameraGet('up')

    @up.setter
    def up(self, value):
        self._plotter._remoteCameraSet('up', value)

    @property
    def clipping_range(self):
        return self._plotter._remoteCameraGet('clipping_range')

    @clipping_range.setter
    def clipping_range(self, value):
        self._plotter._remoteCameraSet('clipping_range', value)

    @property
    def parallel_scale(self):
        return self._plotter._remoteCameraGet('parallel_scale')

    @parallel_scale.setter
    def parallel_scale(self, value):
        self._plotter._remoteCameraSet('parallel_scale', value)

    @property
    def view_angle(self):
        return self._plotter._remoteCameraGet('view_angle')

    @view_angle.setter
    def view_angle(self, value):
        self._plotter._remoteCameraSet('view_angle', value)

    def enable_parallel_projection(self):
        self._plotter._remoteCameraCall('enable_parallel_projection')

    def zoom(self, value):
        self._plotter._remoteCameraCall('zoom', value)


class RemotePlotterProxyBase:
    _camera: RemoteCameraProxy | None = None
    _mapper: RemoteMapper | None = None

    _doQueueCallsAndReturnImmediately: bool = False
    _queuedCalls: list[tuple[str, str, tuple, dict | None, tuple]]

    @attrs.define
    class CallbackRegistry:
        _callbacks: dict[str, tp.Callable] = attrs.field(factory=dict)

        def register(self, func: tp.Callable, key: str | None = None) -> str:
            if key is None:
                key = makeStrUnique(baseStr='Callback',
                                    existingStrs=self._callbacks.keys())
            else:
                assert key not in self._callbacks
            self._callbacks[key] = func
            return key

        def callback(self, name: str, *args, **kwargs):
            return self._callbacks[name](*args, **kwargs)

    _callbackRegistry: CallbackRegistry

    def __init__(self):
        logger.debug(f'Initializing {self.__class__.__name__}')
        self._callbackRegistry = self.CallbackRegistry()

        self._isReady = asyncio.Event()

        self._queuedCalls = []

    @property
    def picked_point(self):
        return self._remotePlotterGet('picked_point')

    @property
    def camera(self):
        if self._camera is None:
            self._remoteQueryProperty('camera')
            self._camera = RemoteCameraProxy(plotter=self)
        return self._camera

    @property
    def mapper(self):
        if self._mapper is None:
            self._mapper = RemoteMapper(parentPlotter=self)
        return self._mapper

    @contextmanager
    def allowNonblockingCalls(self):
        """
        Within this context, most commands will be pushed out to remote for execution, and we won't wait for the return value (i.e. we will immediately return None).

        Some commands may still block.

        Multiple calls are guaranteed to be executed in order by the remote plotter. If multiple non-blocking calls are made right before a blocking call, the blocking call will block until all preceding non-blocking calls are complete before starting.
        """
        prevVal = self._doQueueCallsAndReturnImmediately
        self._doQueueCallsAndReturnImmediately = True
        yield
        self._doQueueCallsAndReturnImmediately = prevVal

    @contextmanager
    def disallowNonblockingCalls(self):
        prevVal = self._doQueueCallsAndReturnImmediately
        self._doQueueCallsAndReturnImmediately = False
        yield
        self._doQueueCallsAndReturnImmediately = prevVal

    async def _sendReqAndRecv_async(self, msg):
        raise NotImplementedError  # to be implemented by subclass

    def _sendReqAndRecv(self, msg):
        raise NotImplementedError  # to be implemented by subclass

    def _sendReqNonblocking(self, msg) -> None:
        raise NotImplementedError  # to be implemented by subclass

    def _prepareForCall(self, cmdKey: str, fnStr: str, args: tuple = (), kwargs: dict | None = None, cmdArgs: tuple = ()):
        if kwargs is None:
            kwargs = dict()

        logger.debug(f'{cmdKey} {fnStr}')
        assert self._isReady.is_set()

        args = list(args)

        for iArg in range(len(args)):
            if isinstance(args[iArg], RemoteActorProxy):
                # convert from RemoteActor to ActorRef
                args[iArg] = ActorRef(actorID=args[iArg].actorID)

        if 'callback' in kwargs:
            # convert from callback function to key matching new entry
            # in callback registry
            callbackFn = kwargs['callback']
            callbackKey = self._callbackRegistry.register(callbackFn)
            kwargs['callback'] = callbackKey

        if 'mesh' in kwargs and isinstance(kwargs['mesh'], pv.PolyData):
            # clear un-pickleable obbTree field
            # note: this may cause unexpected issues...
            kwargs['mesh'] = kwargs['mesh'].copy()
            kwargs['mesh']._obbTree = None

        logger.debug(f'prepared for call {cmdKey} {fnStr}')

        return cmdKey, *cmdArgs, fnStr, args, kwargs

    async def _remoteCall_async(self, cmdKey: str, fnStr: str, args: tuple = (), kwargs: dict | None = None, cmdArgs: tuple = ()):
        req = self._prepareForCall(cmdKey, fnStr, args, kwargs, cmdArgs)

        resp = await self._sendReqAndRecv_async(req)

        return self._handleResp(fnStr, resp)

    def _remoteCall(self, cmdKey: str, fnStr: str, args: tuple = (), kwargs: dict | None = None, cmdArgs: tuple = ()):
        req = self._prepareForCall(cmdKey, fnStr, args, kwargs, cmdArgs)

        if self._doQueueCallsAndReturnImmediately:
            self._sendReqNonblocking(req)
            return None
        else:
            resp = self._sendReqAndRecv(req)
            logger.debug(f'Waiting for response to {fnStr}')
            return self._handleResp(fnStr, resp)

    def _handleResp(self, label, resp):
        logger.debug(f'{label} response: {resp}')
        if isinstance(resp, Exception):
            logger.error(f'{exceptionToStr(resp)}')
            raise resp
        elif isinstance(resp, ActorRef):
            # convert result to a RemoteActor
            return RemoteActorProxy(actorID=resp.actorID, plotter=self)
        else:
            return resp

    async def _handleMsg(self, msg):
        match msg[0]:
            case 'callback':
                callbackKey = msg[1]
                args = msg[2]
                kwargs = msg[3]
                callback = self._callbackRegistry._callbacks[callbackKey]
                resp = callback(*args, **kwargs)
                assert resp is None
                return 'ack'
            case _:
                raise NotImplementedError(f'Unexpected message type: {msg[0]}')

    def _remotePlotterCall(self, fnStr, *args, **kwargs):
        return self._remoteCall('callPlotterMethod', fnStr, args, kwargs)

    async def _remotePlotterCall_async(self, fnStr, *args, **kwargs):
        return await self._remoteCall_async('callPlotterMethod', fnStr, args, kwargs)

    def _remotePlotterGet(self, key: str):
        return self._remoteCall('plotterGet', key)

    def _remoteActorCall(self, actor: RemoteActorProxy, fnStr, *args, **kwargs):
        actorRef = ActorRef(actorID=actor.actorID)

        return self._remoteCall('callActorMethod', fnStr, args, kwargs, cmdArgs=(actorRef,))

    def _remoteActorMapperCall(self, actor: RemoteActorProxy, fnStr: str, *args, **kwargs):
        actorRef = ActorRef(actorID=actor.actorID)
        return self._remoteCall('callActorMapperMethod', fnStr, args, kwargs, cmdArgs=(actorRef,))

    def _remoteActorMapperGet(self, actor: RemoteActorProxy, key: str):
        actorRef = ActorRef(actorID=actor.actorID)
        return self._remoteCall('actorMapperGet', key, cmdArgs=(actorRef,))

    def _remoteActorMapperSet(self, actor: RemoteActorProxy, key: str, value):
        actorRef = ActorRef(actorID=actor.actorID)
        return self._remoteCall('actorMapperSet', key, (value,), cmdArgs=(actorRef,))

    def _remoteMapperCall(self, fnStr: str, *args, **kwargs):
        return self._remoteCall('callMapperMethod', fnStr, args, kwargs)

    def _remoteMapperGet(self, key: str):
        return self._remoteCall('mapperGet', key)

    def _remoteMapperSet(self, key: str, value):
        return self._remoteCall('mapperSet', key, (value,))

    def _remoteCameraGet(self, key: str):
        return self._remoteCall('cameraGet', key)

    def _remoteCameraSet(self, key: str, value):
        return self._remoteCall('cameraSet', key, (value,))

    def _remoteCameraCall(self, fnStr, *args, **kwargs):
        return self._remoteCall('cameraCall', fnStr, args, kwargs)

    def _remoteQueryProperty(self, key: str) -> None:
        return self._remoteCall('queryProperty', key)

    def render(self):
        return self._remotePlotterCall('render')

    def subplot(self, row: int, col: int):
        return self._remotePlotterCall('subplot', row, col)

    def add_mesh(self, *args, **kwargs):
        return self._remotePlotterCall('add_mesh', *args, **kwargs)

    def addMesh(self, *args, **kwargs):
        return self._remotePlotterCall('addMesh', *args, **kwargs)

    def add_volume(self, *args, **kwargs):
        return self._remotePlotterCall('add_volume', *args, **kwargs)

    def add_lines(self, *args, **kwargs):
        return self._remotePlotterCall('add_lines', *args, **kwargs)

    async def add_lines_async(self, *args, **kwargs):
        return await self._remotePlotterCall_async('add_lines', *args, **kwargs)

    def addLineSegments(self, *args, **kwargs):
        return self._remotePlotterCall('addLineSegments', *args, **kwargs)

    def add_points(self, *args, **kwargs):
        return self._remotePlotterCall('add_points', *args, **kwargs)

    async def add_points_async(self, *args, **kwargs):
        return await self._remotePlotterCall_async('add_points', *args, **kwargs)

    def add_point_labels(self, *args, **kwargs):
        return self._remotePlotterCall('add_point_labels', *args, **kwargs)

    def add_axes_marker(self, *args, **kwargs):
        return self._remotePlotterCall('add_axes_marker', *args, **kwargs)

    def addIrenStyleClassObserver(self, *args, **kwargs):
        return self._remotePlotterCall('addIrenStyleClassObserver', *args, **kwargs)

    def clear(self, *args, **kwargs):
        return self._remotePlotterCall('clear', *args, **kwargs)

    def enable_depth_peeling(self, *args, **kwargs):
        return self._remotePlotterCall('enable_depth_peeling', *args, **kwargs)

    def enable_parallel_projection(self):
        self._remotePlotterCall('enable_parallel_projection')

    def enable_point_picking(self, *args, **kwargs):
        self._remotePlotterCall('enable_point_picking', *args, **kwargs)

    def enable_image_style(self, *args, **kwargs):
        self._remotePlotterCall('enable_image_style', *args, **kwargs)

    def set_background(self, *args, **kwargs):
        return self._remotePlotterCall('set_background', *args, **kwargs)

    def set_camera_clipping_range(self, *args, **kwargs):
        return self._remotePlotterCall('set_camera_clipping_range', *args, **kwargs)

    def reset_camera_clipping_range(self, *args, **kwargs):
        return self._remotePlotterCall('reset_camera_clipping_range', *args, **kwargs)

    def reset_scalar_bar_ranges(self, *args, **kwargs):
        return self._remotePlotterCall('reset_scalar_bar_ranges', *args, **kwargs)

    def add_axes_at_origin(self, *args, **kwargs):
        return self._remotePlotterCall('add_axes_at_origin', *args, **kwargs)

    def setActorUserTransform(self, actor: RemoteActorProxy, transform: np.ndarray):
        return self._remotePlotterCall('setActorUserTransform', ActorRef(actorID=actor.actorID), transform)

    def reset_camera(self):
        return self._remotePlotterCall('reset_camera')

    def show_grid(self, *args, **kwargs):
        return self._remotePlotterCall('show_grid', *args, **kwargs)

    def showGrid(self, *args, **kwargs):
        return self._remotePlotterCall('showGrid', *args, **kwargs)

    def remove_actor(self, *args, **kwargs):
        return self._remotePlotterCall('remove_actor', *args, **kwargs)

    def pauseRendering(self):
        return self._remotePlotterCall('pauseRendering')

    def maybeResumeRendering(self):
        return self._remotePlotterCall('maybeResumeRendering')

    def resumeRendering(self):
        return self._remotePlotterCall('resumeRendering')

    @contextmanager
    def renderingPaused(self):
        self.pauseRendering()
        yield
        self.maybeResumeRendering()

    def update(self, *args, **kwargs):
        return self._remotePlotterCall('update', *args, **kwargs)

    def update_scalars(self, *args, **kwargs):
        return self._remotePlotterCall('update_scalars', *args, **kwargs)

    def update_scalar_bar_range(self, *args, **kwargs):
        return self._remotePlotterCall('update_scalar_bar_range', *args, **kwargs)

    def updateScalarBarRangeWithVol(self, *args, **kwargs):
        return self._remotePlotterCall('updateScalarBarRangeWithVol', *args, **kwargs)


class RemotePlotterProxy(RemotePlotterProxyBase, QtWidgets.QWidget):
    """
    There are issues with instantiating multiple pyvista BackgroundPlotters in the same
    process, related to opengl window binding and threaded render contexts. To work around
    this, we create a separate process for each  plotter, embed it in the primary Qt process
    using QWidget.createWindowContainer(), and communicate with it via zmq sockets. This is
     very inefficient (due to multiprocesing overhead), but has the side benefit of
      parallelizing some of the rendering workload.

    Note that this requires serializing / deserializing communication between the main process
    and the plotter. Only a limited subset of plotter methods are currently supported, and
    code working with the results of plotter calls (e.g. actors) may need to be further
     adapted by the caller.
    """

    _RemotePlotterApp: tp.Type[RemotePlotterApp] | None = None

    def __init__(self, parent=None, **kwargs):
        RemotePlotterProxyBase.__init__(self)
        QtWidgets.QWidget.__init__(self, parent=parent)

        self.setSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding,
                           QtWidgets.QSizePolicy.Policy.Expanding)

        ctx = zmq.Context()
        actx = azmq.Context()
        self._repSocket = actx.socket(zmq.REP)
        repPort = self._repSocket.bind_to_random_port('tcp://127.0.0.1')

        self._reqSocket = ctx.socket(zmq.REQ)
        self._reqSocketReqPending: bool = False
        self._areqSocket = actx.socket(zmq.REQ)
        self._pushSocket = ctx.socket(zmq.PUSH)
        # connect these later
        self._areqLock = asyncio.Lock()

        procKwargs = dict(reqPort=repPort, plotterKwargs=kwargs)
        self.remoteProc = None
        self._startRemoteProc(procKwargs, **kwargs)

        self._socketLoopTask = asyncio.create_task(asyncTryAndLogExceptionOnError(self._socketLoop))

    @property
    def isReadyEvent(self):
        return self._isReady

    def _startRemoteProc(self, procKwargs, **kwargs):
        assert self.remoteProc is None

        if self._RemotePlotterApp is None:
            self._RemotePlotterApp = RemotePlotterApp

        if True:
            # set log filepath of remote proc based on filepath of root logger file handler
            handlers = [h for h in logging.getLogger().handlers if isinstance(h, logging.FileHandler)]
            if len(handlers) > 0:
                logFilepath = handlers[-1].baseFilename
                procKwargs = procKwargs.copy()
                procKwargs['logFilepath'] = logFilepath

        procKwargs['theme'] = 'light' if self.palette().color(QtGui.QPalette.Base).value() > 128 else 'dark'
        # TODO: add support for dynamically updating theme on palette changes later

        self.remoteProc = mp.Process(target=self._RemotePlotterApp.createAndRun,
                                     daemon=True,
                                     kwargs=procKwargs)
        logger.debug('Starting remote plotter process')
        self.remoteProc.start()

    async def _sendReqAndRecv_async(self, msg):
        async with self._areqLock:
            logger.debug(f'Sending async msg {msg}')
            self._areqSocket.send_pyobj(msg)
            logger.debug(f'Awaiting reply')
            resp = await self._areqSocket.recv_pyobj()
            logger.debug(f'Received reply {resp}')
            return resp

    def _sendReqAndRecv(self, msg):
        """
        Note: due to some weirdness with Qt window containering, the remote process can
        deadlock if we don't keep processing Qt events in the main process during cursor
        interaction with the remote plotter window. So make sure to keep processing events
        even while waiting for an otherwise blocking result to come back.
        """
        if self._reqSocketReqPending:
            if True:
                logger.error(f'Previous request still pending, not sending this: {msg}')
                return
            else:
                logger.warning('Waiting for a previous request to finish before sending this one')
                # note: if previous request was outside of QEventLoop, we will be stuck here and never get its response
                while self._reqSocketReqPending:
                    logger.debug('Waiting...')
                    QtWidgets.QApplication.instance().processEvents(QtCore.QEventLoop.ProcessEventsFlag.ExcludeUserInputEvents)
                    time.sleep(0.001)

        try:
            logger.debug(f'Sending msg {msg}')
            self._reqSocket.send_pyobj(msg)
        except TypeError as e:
            logger.error(f'Problem serializing message: {exceptionToStr(e)}')
            raise e

        logger.debug('Waiting for response')
        self._reqSocketReqPending = True
        while True:
            try:
                resp = self._reqSocket.recv_pyobj(flags=zmq.NOBLOCK)
            except zmq.error.Again:
                # no message available
                logger.debug('No message available, waiting...')
                QtWidgets.QApplication.instance().processEvents(QtCore.QEventLoop.ProcessEventsFlag.ExcludeUserInputEvents)
                time.sleep(0.001)
            except Exception as e:
                logger.error(f'Unhandled exception in _waitForResp: {exceptionToStr(e)}')
                raise e
            else:
                break
        self._reqSocketReqPending = False
        logger.debug(f'Received reply {resp}')
        return resp

    def _sendReqNonblocking(self, msg):
        self._pushSocket.send_pyobj(msg)
        return None

    async def _socketLoop(self):
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        tempWdgt = QtWidgets.QLabel()
        tempWdgt.setText('Initializing plotter...')
        tempWdgt.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter | QtCore.Qt.AlignmentFlag.AlignVCenter)
        layout.addWidget(tempWdgt)

        logger.debug('Waiting for remote to start up')
        msg = await self._repSocket.recv_pyobj()
        assert isinstance(msg, tuple)
        # first message sent should be remote's repPort and pullPort
        assert msg[0] == 'ports'
        portsDict = msg[1]
        logger.debug(f'Got remote ports: {portsDict}')

        self._repSocket.send_pyobj('ack')

        self._reqSocket.connect(f'tcp://localhost:{portsDict["rep"]}')
        self._areqSocket.connect(f'tcp://localhost:{portsDict["rep"]}')
        self._pushSocket.connect(f'tcp://localhost:{portsDict["pull"]}')

        # get win ID and reparent remote window
        winID = await self._sendReqAndRecv_async(('getWinID',))

        self._embedWin = QtGui.QWindow.fromWinId(winID)
        assert self._embedWin != 0
        #self._embedWin.setFlags(QtCore.Qt.FramelessWindowHint)
        if False:  # TODO: debug, delete or set to False
            self._embedWdgt = QtWidgets.QWidget()  # placeholder
        else:
            try:
                self._embedWdgt = QtWidgets.QWidget.createWindowContainer(self._embedWin, parent=self)
            except RuntimeError as e:
                # if this widget was deleted while awaiting remote initialization, will
                # raise a runtime error here
                logger.warning('Problem while creating window container, giving up')
                self.remoteProc.terminate()
                return

        layout.removeWidget(tempWdgt)
        tempWdgt.deleteLater()
        layout.addWidget(self._embedWdgt)

        resp = await self._sendReqAndRecv_async(('showWindow',))
        assert resp == 'ack'

        self._isReady.set()

        while True:
            msg = await self._repSocket.recv_pyobj()
            try:
                resp = await self._handleMsg(msg)
            except Exception as e:
                logger.error(f'Exception while handling message: {exceptionToStr(e)}')
                resp = e

            await self._repSocket.send_pyobj(resp)

    def render(self, *args, **kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return RemotePlotterProxyBase.render(self)
        else:
            return QtWidgets.QWidget.render(self, *args, **kwargs)

    async def close_async(self):
        logger.info('Closing')
        await self._isReady.wait()
        await self._sendReqAndRecv_async(('quit',))
        self._socketLoopTask.cancel()
        self.remoteProc.terminate()

    def close(self):
        logger.info('Closing')
        asyncio.create_task(asyncTryAndLogExceptionOnError(self.close_async))
        super().close()


@attrs.define
class RemoteMapper:
    _parentActor: RemoteActorProxy | None = None
    _parentPlotter: RemotePlotterProxy | None = None

    def SetInputData(self, *args, **kwargs):
        if self._parentActor is None:
            raise NotImplementedError('Only supported when mapper created from a specific parent actor')

        return self._parentActor.plotter._remoteActorMapperCall(
            self._parentActor,
            'SetInputData', *args, **kwargs)

    @property
    def scalar_range(self):
        if self._parentActor is None:
            return self._parentPlotter._remoteMapperGet('scalar_range')
        else:
            return self._parentActor.plotter._remoteActorMapperGet(self._parentActor, 'scalar_range')

    @scalar_range.setter
    def scalar_range(self, value: tuple):
        if self._parentActor is None:
            self._parentPlotter._remoteMapperSet('scalar_range', value)
        else:
            self._parentActor.plotter._remoteActorMapperSet(self._parentActor, 'scalar_range', value)

    def update(self):
        if self._parentActor is None:
            return self._parentPlotter._remoteMapperCall('update')
        else:
            return self._parentActor.plotter._remoteActorMapperCall(self._parentActor, 'update')


from __future__ import annotations

import asyncio
import logging
import typing as tp
from typing import ClassVar

import attrs
import numpy as np
import pyvista as pv
import vtkmodules.vtkRenderingAnnotation
import zmq
from qtpy import QtWidgets, QtCore
from zmq import asyncio as azmq

from NaviNIBS.util import exceptionToStr
from NaviNIBS.util import ZMQAsyncioFix
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.GUI.QAppWithAsyncioLoop import RunnableAsApp
from NaviNIBS.util.logging import createLogFileHandler
from NaviNIBS.util.pyvista import Actor
from NaviNIBS.util.pyvista.plotting import BackgroundPlotter
from NaviNIBS.util.pyvista.RemotePlotting import ActorRef

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@attrs.define
class ActorManager:
    _counter: int = 0
    _actors: dict[ActorRef, Actor] = attrs.field(factory=dict)

    def addActor(self, actor: Actor) -> ActorRef:
        self._counter += 1
        actorID = f'<Actor{self._counter}>'
        actorRef = ActorRef(actorID)
        self._actors[actorRef] = actor
        return actorRef

    def getActor(self, actorRef: ActorRef) -> Actor:
        return self._actors[actorRef]


@attrs.define(kw_only=True)
class RemotePlotManagerBase:
    _parentLayout: QtWidgets.QLayout
    _plotterArgs: list = attrs.field(factory=list)
    _plotterKwargs: dict = attrs.field(factory=dict)

    _actorManager: ActorManager = attrs.field(init=False, factory=ActorManager)
    _Plotter: ClassVar
    _plotter: RemotePlotter | None = attrs.field(default=None)

    def __attrs_post_init__(self):
        pass

    def initPlotter(self):
        # don't do this automatically to allow deferred widget instantiation
        assert self._plotter is None

        logger.info('Initializing plotter')
        self._plotter = self._Plotter(*self._plotterArgs,
                                       #parent=self._parentLayout.parentWidget(),
                                       **self._plotterKwargs)
        logger.debug('Adding plotter to parent layout')
        self._parentLayout.addWidget(self._plotter)

    @property
    def plotter(self):
        return self._plotter

    async def _handleMsg(self, msg):
        """
        :return: response to message
        """
        if not isinstance(msg, tuple):
            raise NotImplementedError('Unexpected message format')

        match msg[0]:
            case 'noop':
                return 'ack'

            case 'quit':
                asyncio.create_task(asyncTryAndLogExceptionOnError(self._closeAfterDelay))
                return 'ack'

            case 'getWinID':
                if True:
                    win = self._parentLayout.parentWidget().window()
                else:
                    win = self._plotter.window()
                return win.winId()

            case 'showWindow':
                if True:
                    win = self._parentLayout.parentWidget().window()
                else:
                    win = self._plotter.window()

                if self._plotter is None:
                    # assume everything else is now ready and we should instantiate plotter
                    self.initPlotter()

                win.show()
                return 'ack'

            case 'plotterGet':
                assert isinstance(msg[1], str)
                assert len(msg) == 4 and len(msg[2]) == 0 and len(msg[3]) == 0
                return getattr(self._plotter, msg[1])

            case 'callPlotterMethod':
                return self._callPlotMethod(msg[1:])

            case 'callActorMethod':
                return self._callActorMethod(msg[1], msg[2:])

            case 'callActorMapperMethod':
                return self._callMapperMethod(msg[2:], msg[1])

            case 'actorMapperGet':
                assert isinstance(msg[1], ActorRef)
                assert isinstance(msg[2], str)
                assert len(msg) == 5 and len(msg[3]) == 0 and len(msg[4]) == 0
                return getattr(self._actorManager.getActor(msg[1]).GetMapper(), msg[2])

            case 'actorMapperSet':
                assert isinstance(msg[1], ActorRef)
                assert isinstance(msg[2], str)
                assert len(msg) == 5
                assert len(msg[3]) == 1
                assert len(msg[4]) == 0
                setattr(self._actorManager.getActor(msg[1]).GetMapper(), msg[2], msg[3][0])
                return None

            case 'callMapperMethod':
                return self._callMapperMethod(msg[1:], None)

            case 'mapperGet':
                assert isinstance(msg[1], str)
                assert len(msg) == 4 and len(msg[2]) == 0 and len(msg[3]) == 0
                return getattr(self._plotter.mapper, msg[1])

            case 'mapperSet':
                assert isinstance(msg[1], str)
                assert len(msg) == 4
                assert len(msg[2]) == 1
                assert len(msg[3]) == 0
                setattr(self._plotter.mapper, msg[1], msg[2][0])
                return None

            case 'cameraGet':
                assert isinstance(msg[1], str)
                assert len(msg) == 4 and len(msg[2]) == 0 and len(msg[3]) == 0
                return getattr(self._plotter.camera, msg[1])

            case 'cameraSet':
                assert isinstance(msg[1], str)
                assert len(msg) == 4
                assert len(msg[2]) == 1
                assert len(msg[3]) == 0
                setattr(self._plotter.camera, msg[1], msg[2][0])
                return None

            case 'cameraCall':
                return self._callCameraMethod(msg[1:])

            case 'queryProperty':
                assert isinstance(msg[1], str)
                self._queryProperty(msg[1])
                return None

            case _:
                raise NotImplementedError(f'Unexpected message type: {msg[0]}')

    def _queryProperty(self, key: str) -> None:
        """
        Call a getter without returning result.

        Useful in rare circumstances like getting plotter camera to get
        forced camera reset out of the way, without trying to actually
        serialize the camera object
        """
        assert self._plotter is not None
        fn = getattr
        args = (self._plotter, key)
        self._callMethod(fn, args, {})
        return

    def _callPlotMethod(self, msg):
        assert self._plotter is not None
        fn = getattr(self._plotter, msg[0])
        args = list(msg[1])
        kwargs = msg[2]

        return self._callMethod(fn, args, kwargs)

    def _callActorMethod(self, actor: ActorRef, msg):
        actor = self._actorManager.getActor(actor)
        fn = getattr(actor, msg[0])
        args = list(msg[1])
        kwargs = msg[2]

        return self._callMethod(fn, args, kwargs)

    def _callMapperMethod(self, msg, actor: ActorRef | None):
        if actor is not None:
            actor = self._actorManager.getActor(actor)
            mapper = actor.GetMapper()
        else:
            mapper = self._plotter.mapper  # only set by some plotting functions (e.g. add_volume)
        fn = getattr(mapper, msg[0])
        args = list(msg[1])
        kwargs = msg[2]

        return self._callMethod(fn, args, kwargs)

    def _callCameraMethod(self, msg):
        assert self._plotter is not None
        fn = getattr(self._plotter.camera, msg[0])
        args = list(msg[1])
        kwargs = msg[2]

        return self._callMethod(fn, args, kwargs)

    def _callMethod(self, fn, args, kwargs):
        logger.debug(f'calling method {fn} {args} {kwargs}')
        # convert any obvious ActorRefs to Actors
        for iArg in range(len(args)):
            if isinstance(args[iArg], ActorRef):
                args[iArg] = self._actorManager.getActor(args[iArg])
        for key in kwargs:
            if isinstance(kwargs[key], ActorRef):
                kwargs[key] = self._actorManager.getActor(kwargs[key])

        # convert any callback keys in kwargs to callbacks
        if 'callback' in kwargs:
            callbackKey = kwargs['callback']
            callbackFn = lambda *args, **kwargs: self._executeCallback(callbackKey, *args, **kwargs)
            kwargs['callback'] = callbackFn

        # call method
        result = fn(*args, **kwargs)

        # convert any Actors to ActorRefs
        if isinstance(result, (Actor,
                               vtkmodules.vtkRenderingCore.vtkActor2D,
                               vtkmodules.vtkRenderingAnnotation.vtkAxesActor,
                               pv._vtk.vtkVolume)):
            # Actor is not pickleable, so convert to an ActorRef
            actorRef = self._actorManager.addActor(result)
            result = actorRef

        logger.debug(f'result of calling method {fn}: {result}')

        return result

    def _executeCallback(self, callbackKey: str, *args, **kwargs):
        raise NotImplementedError  # to be implemented by subclass

    async def _closeAfterDelay(self):
        await asyncio.sleep(0.1)
        self._parentLayout.parentWidget().window().close()


class RemotePlotterMixin:
    def setActorUserTransform(self, actor: Actor, transform: np.ndarray):
        # convert from numpy array to vtkTransform
        t = pv._vtk.vtkTransform()
        t.SetMatrix(pv.vtkmatrix_from_array(transform))

        actor.SetUserTransform(t)


class RemotePlotter(BackgroundPlotter, RemotePlotterMixin):
    def __init__(self, *args, **kwargs):
        BackgroundPlotter.__init__(self, *args, **kwargs)
        RemotePlotterMixin.__init__(self)


@attrs.define(kw_only=True)
class RemotePlotManager(RemotePlotManagerBase):
    _reqPort: int
    _repPort: int | None = None
    _pullPort: int | None = None
    _addr: str = '127.0.0.1'
    _Plotter: ClassVar = RemotePlotter
    _ctx: azmq.Context = attrs.field(init=False, factory=azmq.Context)
    _reqSock: azmq.Socket = attrs.field(init=False)
    _repSock: azmq.Socket = attrs.field(init=False)
    _pullSock: azmq.Socket = attrs.field(init=False)
    _reqLock: asyncio.Lock = attrs.field(init=False, factory=asyncio.Lock)
    _socketLoopTask: asyncio.Task = attrs.field(init=False)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._reqSock = self._ctx.socket(zmq.REQ)
        self._reqSock.connect(f'tcp://{self._addr}:{self._reqPort}')

        self._repSock = self._ctx.socket(zmq.REP)
        if self._repPort is None:
            self._repPort = self._repSock.bind_to_random_port(f'tcp://{self._addr}')
        else:
            self._repSock.bind(f'tcp://{self._addr}:{self._repPort}')

        self._pullSock = self._ctx.socket(zmq.PULL)
        if self._pullPort is None:
            self._pullPort = self._pullSock.bind_to_random_port(f'tcp://{self._addr}')
        else:
            self._pullSock.bind(f'tcp://{self._addr}:{self._pullPort}')

        self._socketLoopTask = asyncio.create_task(asyncTryAndLogExceptionOnError(self._socketLoop))

    async def _socketLoop(self):

        async with self._reqLock:
            await self._reqSock.send_pyobj(('ports', dict(rep=self._repPort,
                                                          pull=self._pullPort)))
            resp = await self._reqSock.recv_pyobj()
            assert resp == 'ack'

        if True:
            self.initPlotter()

        poller = azmq.Poller()
        poller.register(self._pullSock, zmq.POLLIN)
        poller.register(self._repSock, zmq.POLLIN)
        while True:
            logger.debug('Awaiting msg')
            socks = dict(await poller.poll())
            with self.plotter.renderingPaused():
                while len(socks) > 0:
                    if self._pullSock in socks:
                        # note: this ordering has the effect of emptying pull queue before checking rep
                        # queue, which is what we want to deplete all pending non-blocking requests before
                        # we respond to a blocking request (which was presumably sent later by
                        # the single client)
                        msg = await self._pullSock.recv_pyobj()
                        replyOnSock = None  # push-pull is unidirectional (nonblocking command, no return)
                    elif self._repSock in socks:
                        msg = await self._repSock.recv_pyobj()
                        replyOnSock = self._repSock
                    else:
                        raise NotImplementedError

                    logger.debug(f'Received msg: {msg}')

                    try:
                        resp = await self._handleMsg(msg)
                    except Exception as e:
                        logger.error(f'Exception while handling msg {msg}: {exceptionToStr(e)}')
                        resp = e

                    if replyOnSock is not None:
                        logger.debug(f'Sending response: {resp}')
                        await replyOnSock.send_pyobj(resp)
                    else:
                        logger.debug(f'Non-blocking request complete, dropping response: {resp}')

                    if True:
                        # wait to render until after processed all pending requests
                        socks = dict(await poller.poll(timeout=0))
                    else:
                        socks = dict()

    def _executeCallback(self, callbackKey: str, *args, **kwargs):
        logger.debug(f'Queuing async task for callback {callbackKey}')
        asyncio.create_task(asyncTryAndLogExceptionOnError(self._executeCallbackAsync, callbackKey, *args, **kwargs))

    async def _executeCallbackAsync(self, callbackKey: str, *args, **kwargs):
        async with self._reqLock:
            await self._reqSock.send_pyobj(('callback', callbackKey, args, kwargs))
            resp = await self._reqSock.recv_pyobj()
            assert resp == 'ack'


@attrs.define(kw_only=True)
class RemotePlotterApp(RunnableAsApp):
    _reqPort: int
    _repPort: int | None = None
    _appName: str = 'RemotePlotter'
    _logFilepath: str | None = None
    _plotterKwargs: dict = attrs.field(factory=dict)

    _plotManager: RemotePlotManager = attrs.field(init=False, default=None)
    _rootWdgt: QtWidgets.QWidget = attrs.field(init=False)
    _debugTimer: QtCore.QTimer = attrs.field(init=False)
    _logFileHandler: logging.FileHandler = attrs.field(init=False)

    def __attrs_post_init__(self):
        if self._logFilepath is not None:
            self._logFileHandler = createLogFileHandler(self._logFilepath)
            logging.getLogger('').addHandler(self._logFileHandler)

        logger.debug(f'Initializing {self.__class__.__name__}')
        super().__attrs_post_init__()
        wdgt = QtWidgets.QWidget()
        self._rootWdgt = wdgt
        wdgt.setLayout(QtWidgets.QVBoxLayout())
        wdgt.layout().setContentsMargins(0, 0, 0, 0)

        self._win.setCentralWidget(wdgt)

        flags = self._win.windowFlags()
        flags |= QtCore.Qt.FramelessWindowHint
        flags |= QtCore.Qt.MSWindowsFixedSizeDialogHint
        flags |= QtCore.Qt.SubWindow
        self._win.setWindowFlags(flags)

        self._initPlotManager()

    def _initPlotManager(self):
        assert self._plotManager is None
        self._plotManager = RemotePlotManager(reqPort=self._reqPort,
                                              repPort=self._repPort,
                                              parentLayout=self._rootWdgt.layout(),
                                              plotterKwargs=self._plotterKwargs)



from __future__ import annotations

import asyncio

import attrs
from datetime import datetime
import logging
import numpy as np
import os
import pathlib
import pyvista as pv
import qtawesome as qta
from qtpy import QtWidgets, QtGui, QtCore
import shutil
import typing as tp

from . import MainViewPanel
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.Navigator.GUI.Widgets.MRIViews import MRISliceView
from NaviNIBS.Navigator.GUI.Widgets.SurfViews import Surf3DView
from NaviNIBS.Navigator.GUI.Widgets.CollectionTableWidget import PlanningFiducialsTableWidget
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.GUI.Icons import getIcon
from NaviNIBS.util.GUI.QFileSelectWidget import QFileSelectWidget
from NaviNIBS.util.pyvista import RemotePlotterProxy
from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.Navigator.Model.SubjectRegistration import Fiducial


logger = logging.getLogger(__name__)


@attrs.define
class FiducialsPanel(MainViewPanel):
    _key: str = 'Plan fiducials'
    _icon: QtGui.QIcon = attrs.field(init=False, factory=lambda: getIcon('mdi6.head-snowflake-outline'))
    _tblWdgt: PlanningFiducialsTableWidget = attrs.field(init=False)
    _views: tp.Dict[str, tp.Union[MRISliceView, Surf3DView]] = attrs.field(init=False, factory=dict)
    _surfKey: str = 'skinSurf'
    _fiducialActors: tp.Dict[str, tp.Any] = attrs.field(init=False, factory=dict)

    finishedAsyncInit: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def canBeEnabled(self) -> tuple[bool, str | None]:
        if self.session is None:
            return False, 'No session set'
        if not self.session.MRI.isSet:
            return False, 'No MRI set'
        if not self.session.headModel.isSet:
            return False, 'No head model set'
        return True, None

    def _finishInitialization(self):
        # don't initialize computationally-demanding views until panel is activated (viewed)

        super()._finishInitialization()

        self._wdgt.setLayout(QtWidgets.QHBoxLayout())

        container = QtWidgets.QGroupBox('Planned fiducials')
        container.setLayout(QtWidgets.QVBoxLayout())
        self._wdgt.layout().addWidget(container)
        container.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.MinimumExpanding)

        btnContainer = QtWidgets.QWidget()
        btnContainer.setLayout(QtWidgets.QGridLayout())
        container.layout().addWidget(btnContainer)

        btn = QtWidgets.QPushButton('Autoset fiducials from head model')
        btn.clicked.connect(self._onAutosetBtnClicked)
        btnContainer.layout().addWidget(btn, 0, 0, 1, 2)

        btn = QtWidgets.QPushButton('Delete fiducial')
        btn.clicked.connect(self._onDeleteBtnClicked)
        btnContainer.layout().addWidget(btn, 1, 1)

        btn = QtWidgets.QPushButton('Set fiducial')
        btn.clicked.connect(self._onSetBtnClicked)
        btnContainer.layout().addWidget(btn, 2, 0)

        btn = QtWidgets.QPushButton('Goto fiducial')
        btn.clicked.connect(self._onGotoBtnClicked)
        btnContainer.layout().addWidget(btn, 2, 1)

        self._tblWdgt = PlanningFiducialsTableWidget()
        self._tblWdgt.sigSelectionChanged.connect(self._onSelectedFiducialsChanged)
        container.layout().addWidget(self._tblWdgt.wdgt)

        container = QtWidgets.QWidget()
        container.setLayout(QtWidgets.QGridLayout())
        self._wdgt.layout().addWidget(container)
        for iRow, iCol, key in ((0, 1, 'x'), (0, 0, 'y'), (1, 0, 'z'), (1, 1, '3D')):
            if key in ('x', 'y', 'z'):
                self._views[key] = MRISliceView(normal=key)
            elif key == '3D':
                self._views[key] = Surf3DView(normal=key, activeSurf=self._surfKey, surfOpacity=0.7)
            else:
                raise NotImplementedError()

            self._views[key].sigSliceOriginChanged.connect(lambda key=key: self._onSliceOriginChanged(sourceKey=key))

            container.layout().addWidget(self._views[key].wdgt, iRow, iCol)

        for key, view in self._views.items():
            if view.session is None and self.session is not None:
                view.session = self.session

        if self.session is not None:
            self._onPanelInitializedAndSessionSet()

        self._onPlannedFiducialsChanged()

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._finishInitialization_async))

    async def _finishInitialization_async(self):
        for viewKey, view in self._views.items():
            if isinstance(view.plotter, RemotePlotterProxy):
                await view.plotter.isReadyEvent.wait()

        self._onPlannedFiducialsChanged()  # update plot

        self.finishedAsyncInit.set()

    def _onSliceOriginChanged(self, sourceKey: str):
        for key, view in self._views.items():
            if key == sourceKey:
                continue
            view.sliceOrigin = self._views[sourceKey].sliceOrigin

    def _onSessionSet(self):
        super()._onSessionSet()

        if self._hasInitialized:
            self._onPanelInitializedAndSessionSet()

    def _onPanelInitializedAndSessionSet(self):
        self.session.subjectRegistration.fiducials.sigItemsChanged.connect(self._onFiducialsChanged)
        self.session.headModel.sigDataChanged.connect(self._onHeadModelUpdated)

        self._tblWdgt.session = self.session

        for key, view in self._views.items():
            view.session = self.session

        self._onPlannedFiducialsChanged()

    def _onTblCellDoubleClicked(self, row: int, col: int):
        coord = list(self.session.subjectRegistration.fiducials.values())[row].plannedCoord
        if coord is not None:
            self._views['3D'].sliceOrigin = coord

    def _onSelectedFiducialsChanged(self, selKeys: list[str]):
        pass

    def _onFiducialsChanged(self, fidKeys: list[str], attribs: tp.Optional[list[str]] = None):
        if attribs is None or 'plannedCoord' in attribs:
            self._onPlannedFiducialsChanged()

    def _onPlannedFiducialsChanged(self):
        logger.debug('Planned fiducials changed. Updating plots')

        for viewKey, view in self._views.items():
            if isinstance(view.plotter, RemotePlotterProxy) and not view.plotter.isReadyEvent.is_set():
                # plotter not ready yet
                return

        labels = []
        coords = np.full((len(self.session.subjectRegistration.fiducials), 3), np.nan)
        for iFid, (label, fid) in enumerate(self.session.subjectRegistration.fiducials.items()):
            coord = fid.plannedCoord
            labels.append(label)
            if coord is not None:
                coords[iFid, :] = coord

        if self._hasInitialized or self._isInitializing:
            for viewKey, view in self._views.items():
                if len(coords) == 0:
                    if viewKey in self._fiducialActors:
                        view.plotter.remove_actor(self._fiducialActors.pop(viewKey), reset_camera=False, render=False)
                    continue

                if viewKey == '3D':
                    self._fiducialActors[viewKey] = view.plotter.add_point_labels(
                        name='plannedFiducials',
                        points=coords,
                        labels=labels,
                        point_color='blue',
                        text_color='blue',
                        point_size=15,
                        font_size=15,
                        shape=None,
                        always_visible=True,
                        render_points_as_spheres=True,
                        reset_camera=False,
                        render=False
                    )
                else:
                    self._fiducialActors[viewKey] = view.plotter.add_points(
                        name='plannedFiducials',
                        points=coords,
                        color='blue',
                        opacity=0.85,
                        point_size=15,
                        render_points_as_spheres=True,
                        reset_camera=False,
                        render=False
                    )

                view.updateView()
        else:
            logger.debug('Not yet initialized/ing, skipping fiducials update')

    def _onHeadModelUpdated(self, whatChanged: str):
        if whatChanged is not None and whatChanged != self._surfKey:
            return
        # individual view handles re-rendering on update, so don't need to do anything here
        pass

    def _onAutosetBtnClicked(self, checked: bool):
        eegPositions = self.session.headModel.eegPositions
        subReg = self.session.subjectRegistration
        if eegPositions is None:
            raise ValueError('No EEG positions available in head model')

        labelMapping = {'LPA': 'LPA', 'NAS': 'Nz', 'RPA': 'RPA'}

        coords = np.zeros((3, 3))
        for iLabel, (label, altLabel) in enumerate(labelMapping.items()):
            coords[iLabel, :] = eegPositions.loc[altLabel, ['x', 'y', 'z']].values
            if label in subReg.fiducials:
                subReg.fiducials[label].plannedCoord = coords[iLabel, :]
            else:
                subReg.fiducials[label] = Fiducial(key=label,
                                                   plannedCoord=coords[iLabel, :])

        if False:
            # also set approximate nose tip
            downDir = -1 * np.cross(coords[2, :] - coords[0, :], coords[1, :] - coords[0, :])
            downDir /= np.linalg.norm(downDir)
            centerToNoseDir = coords[1, :] + downDir * 20 - (coords[2, :] + coords[0, :]) / 2
            centerToNoseDir /= np.linalg.norm(centerToNoseDir)
            projPts = np.dot(getattr(self.session.headModel, self._surfKey).points, centerToNoseDir)
            iMax = np.argmax(projPts)
            noseCoord = getattr(self.session.headModel, self._surfKey).points[iMax, :]
            label = 'NoseTip'
            if label in subReg.fiducials:
                subReg.fiducials[label].plannedCoord = noseCoord
            else:
                subReg.fiducials[label] = Fiducial(key=label,
                                                   plannedCoord=noseCoord)

        # note: any pre-existing fiducials with non-standard names will remain unchanged

        self._tblWdgt.resizeColumnsToContents()

    def _getTblCurrentFiducialKey(self) -> tp.Optional[str]:
        return self._tblWdgt.currentCollectionItemKey

    def _onAddBtnClicked(self, checked: bool):
        raise NotImplementedError()  # TODO

    def _onDeleteBtnClicked(self, checked: bool):
        key = self._getTblCurrentFiducialKey()
        if key is None:
            # no fiducial selected
            return
        logger.info('Deleting {} fiducial'.format(key))
        self.session.subjectRegistration.fiducials.deleteItem(key)

    def _onSetBtnClicked(self, checked: bool):
        key = self._getTblCurrentFiducialKey()
        if key is None:
            # no fiducial selected
            return
        coord = self._views['3D'].sliceOrigin
        logger.info('Setting {} fiducial coord to {}'.format(key, coord))
        self.session.subjectRegistration.fiducials[key].plannedCoord = coord

    def _onGotoBtnClicked(self, checked: bool):
        key = self._getTblCurrentFiducialKey()
        if key is None:
            # no fiducial selected
            return
        coord = self.session.subjectRegistration.fiducials[key].plannedCoord
        if coord is None:
            # no coordinates set for this fiducial
            return
        logger.info('Going to {} at {}'.format(key, coord))
        self._views['3D'].sliceOrigin = coord



from __future__ import annotations

import multiprocessing as mp

import attrs

from NaviNIBS.util.pyvista.RemotePlotting.RemotePlotterProxy import RemotePlotterProxy, RemotePlotterProxyBase
from NaviNIBS.util.pyvista.RemotePlotting.RemoteLayeredPlotter import RemoteLayeredPlotterApp


class RemoteSecondaryLayeredPlotterProxy(RemotePlotterProxyBase):
    _mainPlotter: RemotePrimaryLayeredPlotterProxy
    _layerKey: str
    _rendererLayer: int

    def __init__(self,
                 mainPlotter: RemotePrimaryLayeredPlotterProxy,
                 layerKey: str,
                 rendererLayer: int):
        RemotePlotterProxyBase.__init__(self)
        self._layerKey = layerKey
        self._mainPlotter = mainPlotter
        self._rendererLayer = rendererLayer

        self._isReady.set()  # no async init needed for secondary plotter

    @property
    def rendererLayer(self):
        return self._rendererLayer

    async def _sendReqAndRecv_async(self, msg):
        return await self._mainPlotter._sendReqAndRecv_async(msg, layerKey=self._layerKey)

    def _sendReqAndRecv(self, msg):
        return self._mainPlotter._sendReqAndRecv(msg, layerKey=self._layerKey)

    def _sendReqNonblocking(self, msg) -> None:
        self._mainPlotter._sendReqNonblocking(msg, layerKey=self._layerKey)


class RemotePrimaryLayeredPlotterProxy(RemotePlotterProxy):

    _secondaryPlotters: dict[str, RemoteSecondaryLayeredPlotterProxy]

    def __init__(self, **kwargs):
        RemotePlotterProxy.__init__(self, **kwargs)

        self._secondaryPlotters = dict()

    @property
    def secondaryPlotters(self):
        return self._secondaryPlotters

    def _startRemoteProc(self, procKwargs, **kwargs):

        if self._RemotePlotterApp is None:
            self._RemotePlotterApp = RemoteLayeredPlotterApp

        return super()._startRemoteProc(procKwargs, **kwargs)

    async def _sendReqAndRecv_async(self, msg, layerKey: str | None = '__primary'):
        # add plotter layer key to msg
        msg = (layerKey, *msg)
        return await super()._sendReqAndRecv_async(msg)

    def _sendReqAndRecv(self, msg, layerKey: str | None = '__primary'):
        # add plotter layer key to msg
        msg = (layerKey, *msg)
        return super()._sendReqAndRecv(msg)

    def _sendReqNonblocking(self, msg, layerKey: str | None = '__primary'):
        msg = (layerKey, *msg)
        return super()._sendReqNonblocking(msg)

    def addLayeredPlotter(self, key: str, layer: int | None = None) -> RemoteSecondaryLayeredPlotterProxy:

        assert key not in self._secondaryPlotters
        # add as a remote plotter
        layer = self._remotePlotterCall('addLayeredPlotter', key=key, layer=layer)
        # add as a proxy
        self._secondaryPlotters[key] = RemoteSecondaryLayeredPlotterProxy(
            mainPlotter=self,
            layerKey=key,
            rendererLayer=layer)

        return self._secondaryPlotters[key]

    def setLayer(self, layer: int):
        return self._remotePlotterCall('setLayer', layer=layer)



from __future__ import annotations

import asyncio

import attrs
import logging
import numpy as np
import typing as tp
from typing import ClassVar

from . import PlotViewLayer
from NaviNIBS.util.pyvista import Actor, setActorUserTransform, concatenateLineSegments
from NaviNIBS.util.Transforms import concatenateTransforms, invertTransform


logger = logging.getLogger(__name__)


Transform = np.ndarray


@attrs.define(kw_only=True)
class ToolMeshSurfaceLayer(PlotViewLayer):
    _type: ClassVar[str] = 'ToolMeshSurface'
    _color: str | None = None
    """
    If None, will use tool's defined color
    """
    _toolKey: str
    _doShowToolMesh: bool = True
    _doShowTrackerMesh: bool = False
    _opacity: float | None = None
    """
    If None, will use tool's defined opacity
    """

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._coordinator.positionsClient.sigLatestPositionsChanged.connect(self._onLatestPositionsChanged)

    def _onLatestPositionsChanged(self):
        self._queueRedraw(which='updatePosition')
        # TODO: connect to signals to redraw mesh when tool mesh, color, or opacity changes

    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str, ...]] = None):
        super()._redraw(which=which)

        if not isinstance(which, str):
            # assume parent call above triggered appropriate redraws
            return

        if which == 'all':
            which = ['initSurfs']
            self._redraw(which=which)
            return

        if which == 'initSurfs':
            for toolOrTracker in ('tool', 'tracker'):
                match toolOrTracker:
                    case 'tool':
                        if not self._doShowToolMesh:
                            continue

                    case 'tracker':
                        if not self._doShowTrackerMesh:
                            continue

                    case _:
                        raise NotImplementedError

                tool = self._coordinator.session.tools[self._toolKey]

                mesh = getattr(tool, f'{toolOrTracker}Surf')

                assert mesh is not None

                actorKey = self._getActorKey(f'{toolOrTracker}Surf')

                if actorKey in self._actors:
                    with self._plotter.allowNonblockingCalls():
                        self._plotter.remove_actor(self._actors.pop(actorKey))

                color = self._color
                if color is None:
                    color = getattr(tool, f'{toolOrTracker}Color')
                    if color is None:
                        color = '#999999'  # default color

                opacity = self._opacity
                if opacity is None:
                    opacity = getattr(tool, f'{toolOrTracker}Opacity')
                    if opacity is None:
                        opacity = 1.  # default opacity

                self._actors[actorKey] = self._plotter.add_mesh(mesh=mesh,
                                                                color=color,
                                                                opacity=opacity,
                                                                specular=0.5,
                                                                diffuse=0.5,
                                                                ambient=0.5,
                                                                smooth_shading=True,
                                                                split_sharp_edges=True,
                                                                name=actorKey)

            self._plotter.reset_camera_clipping_range()

            self._redraw('updatePosition')

        elif which == 'updatePosition':

            for toolOrTracker in ('tool', 'tracker'):
                match toolOrTracker:
                    case 'tool':
                        if not self._doShowToolMesh:
                            continue

                    case 'tracker':
                        if not self._doShowTrackerMesh:
                            continue

                    case _:
                        raise NotImplementedError

                actorKey = self._getActorKey(f'{toolOrTracker}Surf')

                try:
                    actor = self._actors[actorKey]
                except KeyError:
                    self._redraw(which='initSurfs')
                    return

                tool = self._coordinator.session.tools[self._toolKey]

                match toolOrTracker:
                    case 'tool':
                        toolOrTrackerStlToTrackerTransf = tool.toolToTrackerTransf @ tool.toolStlToToolTransf
                    case 'tracker':
                        toolOrTrackerStlToTrackerTransf = tool.trackerStlToTrackerTransf
                    case _:
                        raise NotImplementedError

                doHide = True

                match self._plotInSpace:
                    case 'MRI':
                        trackerToWorldTransf = self._coordinator.positionsClient.getLatestTransf(key=tool.trackerKey, default=None)
                        if trackerToWorldTransf is not None:
                            subjectTrackerToWorldTransf = self._coordinator.positionsClient.getLatestTransf(self._coordinator.session.tools.subjectTracker.trackerKey, None)
                            if subjectTrackerToWorldTransf is not None:
                                subjectTrackerToMRITransf = self._coordinator.session.subjectRegistration.trackerToMRITransf
                                if subjectTrackerToMRITransf is not None:
                                    # we have enough info to assemble valid transf
                                    doHide = False

                                    surfToMRITransf = concatenateTransforms([
                                        toolOrTrackerStlToTrackerTransf,
                                        trackerToWorldTransf,
                                        invertTransform(subjectTrackerToWorldTransf),
                                        subjectTrackerToMRITransf,
                                    ])

                                    with self._plotter.allowNonblockingCalls():
                                        setActorUserTransform(actor, surfToMRITransf)
                                        if not actor.GetVisibility():
                                            actor.SetVisibility(True)
                                        self._plotter.render()

                    case 'World':
                        trackerToWorldTransf = self._coordinator.positionsClient.getLatestTransf(key=tool.trackerKey, default=None)
                        if trackerToWorldTransf is not None:
                            doHide = False

                            surfToWorldTransf = concatenateTransforms([
                                toolOrTrackerStlToTrackerTransf,
                                trackerToWorldTransf,
                            ])

                            with self._plotter.allowNonblockingCalls():
                                setActorUserTransform(actor, surfToWorldTransf)
                                if not actor.GetVisibility():
                                    actor.SetVisibility(True)
                                self._plotter.render()

                    case _:
                        raise NotImplementedError

                if doHide:
                    with self._plotter.allowNonblockingCalls():
                        if actor.GetVisibility():
                            actor.SetVisibility(False)
                        self._plotter.render()

        else:
            raise NotImplementedError


@attrs.define
class HeadMeshSurfaceLayer(PlotViewLayer):
    _type: ClassVar[str] = 'HeadMeshSurface'
    _color: str = '#d9a5b2'
    _surfKey: str = 'gmSimpleSurf'
    _opacity: float = 1.

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        if self._plotInSpace != 'MRI':
            self._coordinator.positionsClient.sigLatestPositionsChanged.connect(lambda: self._queueRedraw(which='updatePosition'))

    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str, ...]] = None):
        super()._redraw(which=which)

        if not isinstance(which, str):
            # assume parent call above triggered appropriate redraws
            return

        if which == 'all':
            which = ['initSurf']
            self._redraw(which=which)
            return

        if which == 'initSurf':
            mesh = getattr(self._coordinator.session.headModel, self._surfKey)

            actorKey = self._getActorKey('surf')

            self._actors[actorKey] = self._plotter.add_mesh(mesh=mesh,
                                                            color=self._color,
                                                            opacity=self._opacity,
                                                            specular=0.5,
                                                            diffuse=0.5,
                                                            ambient=0.5,
                                                            smooth_shading=True,
                                                            split_sharp_edges=True,
                                                            name=actorKey)

            self._plotter.reset_camera_clipping_range()

            self._redraw('updatePosition')

        elif which == 'updatePosition':
            match self._plotInSpace:
                case 'MRI':
                    if False:
                        actorKey = self._getActorKey('surf')
                        transf = np.eye(4)
                        setActorUserTransform(self._actors[actorKey], transf)
                        self._plotter.render()
                    else:
                        pass  # assume since plotInSpace is MRI that we don't need to update anything
                case 'World':

                    actorKey = self._getActorKey('surf')
                    if actorKey not in self._actors:
                        self._redraw(which='initSurf')
                        return

                    actor = self._actors[actorKey]

                    doHide = True
                    subjectTrackerToWorldTransf = self._coordinator.positionsClient.getLatestTransf(self._coordinator.session.tools.subjectTracker.trackerKey, None)
                    if subjectTrackerToWorldTransf is not None:
                        subjectTrackerToMRITransf = self._coordinator.session.subjectRegistration.trackerToMRITransf
                        if subjectTrackerToMRITransf is not None:
                            # we have enough info to assemble valid transf
                            doHide = False

                            surfToWorldTransf = concatenateTransforms([
                                invertTransform(subjectTrackerToMRITransf),
                                subjectTrackerToWorldTransf,
                            ])

                            with self._plotter.allowNonblockingCalls():
                                setActorUserTransform(actor, surfToWorldTransf)
                                if not actor.GetVisibility():
                                    actor.SetVisibility(True)
                                self._plotter.render()

                    if doHide:
                        with self._plotter.allowNonblockingCalls():
                            if actor.GetVisibility():
                                actor.SetVisibility(False)
                                self._plotter.render()

                case _:
                    raise NotImplementedError

        else:
            raise NotImplementedError



import asyncio
import attrs
import logging
import numpy as np
import time
import typing as tp
import zmq
import zmq.asyncio as azmq

from NaviNIBS.Devices import positionsServerHostname, positionsServerPubPort, positionsServerCmdPort, TimestampedToolPosition
from NaviNIBS.util import ZMQAsyncioFix
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.ZMQConnector import ZMQConnectorClient, logger as logger_ZMQConnector
from NaviNIBS.util.numpy import array_equalish
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.Transforms import concatenateTransforms
from NaviNIBS.util import exceptionToStr


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger_ZMQConnector.setLevel(logging.INFO)


_novalue = object()


@attrs.define
class ToolPositionsClient:
    _serverHostname: str = positionsServerHostname
    _serverPubPort: int = positionsServerPubPort
    _serverCmdPort: int = positionsServerCmdPort

    _latestPositions: dict[str, tp.Optional[TimestampedToolPosition]] | None = attrs.field(init=False, default=None, repr=False)

    _subSocket: azmq.Socket = attrs.field(init=False)
    _connector: ZMQConnectorClient = attrs.field(init=False, repr=False)

    sigLatestPositionsChanged: Signal = attrs.field(init=False, factory=Signal, repr=False)

    _timeLastHeardFromServer: tp.Optional[float] = None
    _isConnected: bool = False
    _serverStatusTimeout: float = 10.  # consider server offline if we haven't heard from it for this long

    _pollTask: asyncio.Task | None = attrs.field(init=False, repr=False)
    _monitorTask: asyncio.Task | None = attrs.field(init=False, repr=False)
    _startupTask: asyncio.Task | None = attrs.field(init=False, repr=False)

    sigIsConnectedChanged: Signal = attrs.field(init=False, factory=Signal, repr=False)

    def __attrs_post_init__(self):
        ctx = azmq.Context()
        self._subSocket = ctx.socket(zmq.SUB)
        logger.debug('Connecting {}:{}'.format(self._serverHostname, self._serverPubPort))
        self._subSocket.setsockopt(zmq.CONFLATE, 1)
        self._subSocket.connect('tcp://{}:{}'.format(self._serverHostname, self._serverPubPort))
        self._subSocket.setsockopt(zmq.SUBSCRIBE, b'')

        self._connector = ZMQConnectorClient(reqRepPort=self._serverCmdPort,
                                             connAddr=self._serverHostname,
                                             allowAsyncCalls=True)

        self._startupTask = asyncio.create_task(asyncTryAndLogExceptionOnError(
            self._requestLatestPositionsOnStart))

        self._pollTask = asyncio.create_task(asyncTryAndLogExceptionOnError(
            self._receiveLatestPositionsLoop))

        self._monitorTask = asyncio.create_task(asyncTryAndLogExceptionOnError(
            self._monitorServerStatus))

    @property
    def latestPositions(self):
        """
        Note that returned positions may be absolute (rel to world) or relative, based on pos.relativeTo
        """
        return self._latestPositions if self._latestPositions is not None else dict()

    def getServerType(self) -> str:
        return self._connector.get('type')

    def stopReceivingPositions(self):
        if self._pollTask is not None:
            self._pollTask.cancel()
            self._pollTask = None

    def requestLatestPositions(self):
        self._connector.call('publishLatestPositions')

    async def _requestLatestPositionsOnStart(self):
        """
        Whenever (re)connecting to the server, request that all latest positions be re-published
        """
        connectedChangedEvent = asyncio.Event()
        self.sigIsConnectedChanged.connect(lambda: connectedChangedEvent.set())
        while True:
            await connectedChangedEvent.wait()
            connectedChangedEvent.clear()
            if self.isConnected:
                self.requestLatestPositions()

    def getLatestTransf(self, key: str, default: tp.Any = _novalue) -> tp.Optional[np.ndarray]:
        """
        Note that returned transf is always absolute, even if the underlying latest position was relative

        If key is not found, raises KeyError unless a default value is provided (default can be set to None).
        """
        tsPos = self.latestPositions.get(key, None)
        if tsPos is None or tsPos.transf is None:
            if default is _novalue:
                raise KeyError('No matching, valid transf found')
            else:
                return default
        if tsPos.relativeTo != 'world':
            # convert relative transform to world transform
            otherTransf = self.getLatestTransf(key=tsPos.relativeTo, default=None)
            if otherTransf is None:
                return default
            else:
                return concatenateTransforms((tsPos.transf, otherTransf))
        return tsPos.transf

    async def recordNewPosition_async(self, key: str, position: TimestampedToolPosition):
        """
        This should only be used to record positions of tools that are not tracked by the camera
        (e.g. when a position is reported by some other external system, or simulated)
        """
        logger.debug(f'recordNewPosition {key} {position}')
        await self._connector.callAsync_async('recordNewPosition',
                                              key=key,
                                              position=position.asDict())

    def recordNewPosition_sync(self, key: str, position: TimestampedToolPosition):
        """
        This should only be used to record positions of tools that are not tracked by the camera
        (e.g. when a position is reported by some other external system, or simulated)
        """
        logger.debug(f'recordNewPosition {key} {position}')
        self._connector.callAsync('recordNewPosition',
                                  key=key,
                                  position=position.asDict())

    async def _receiveLatestPositionsLoop(self):
        poller = azmq.Poller()
        poller.register(self._subSocket, zmq.POLLIN)
        while True:
            socks = dict(await poller.poll())
            if self._subSocket in socks:
                hasPendingMessages = True
                msg = None
                while hasPendingMessages:
                    msg = await self._subSocket.recv_json()
                    logger.debug('Received published message')
                    #hasPendingMessages = (await self._subSocket.poll(timeout=0.)) > 0
                    hasPendingMessages = False

                self._timeLastHeardFromServer = time.time()
                self._updateIsConnected()

                newPositions = {key: (TimestampedToolPosition.fromDict(val) if val is not None else None) for key, val in msg.items()}

                # only notify if tool position (not time) changed
                positionsChanged = False
                if self._latestPositions is None:
                    # latest positions had never been set, so send out an update
                    # even if empty
                    positionsChanged = True
                else:
                    for key, newPos in newPositions.items():
                        try:
                            oldPos = self._latestPositions[key]
                        except KeyError:
                            positionsChanged = True
                            break
                        else:
                            if not array_equalish(oldPos.transf, newPos.transf) or oldPos.relativeTo != newPos.relativeTo:
                                positionsChanged = True
                                break

                if not positionsChanged:
                    for key, val in self._latestPositions.items():
                        if key not in newPositions:
                            positionsChanged = True
                            break

                if not positionsChanged:
                    logger.debug('Positions not changed during update, not signaling.')
                    continue

                self._latestPositions = {key: (TimestampedToolPosition.fromDict(val) if val is not None else None) for key, val in msg.items()}
                logger.debug('Signaling change in latest positions')
                try:
                    self.sigLatestPositionsChanged.emit()  # only emit for latest in series of updates to avoid falling behind
                except Exception as e:
                    logger.error('Exception during position update:\n {}'.format(exceptionToStr(e)))
                    raise e

    async def _monitorServerStatus(self):
        while True:
            if self._timeLastHeardFromServer is None or (time.time() - self._timeLastHeardFromServer) > self._serverStatusTimeout / 2:
                self._timeLastHeardFromServer = None
                try:
                    await self._connector.ping_async(timeout=self._serverStatusTimeout * 1000. / 2)
                except TimeoutError:
                     logger.debug('Ping to tool positions server timed out.')
                else:
                    self._timeLastHeardFromServer = time.time()
                self._updateIsConnected()
            await asyncio.sleep(self._serverStatusTimeout / 4)

    def _updateIsConnected(self):
        isConnected = self._timeLastHeardFromServer is not None
        if isConnected != self._isConnected:
            self._isConnected = isConnected
            self.sigIsConnectedChanged.emit()

    @property
    def isConnected(self):
        return self._timeLastHeardFromServer is not None

    @classmethod
    async def createAndRun_async(cls, *args, **kwargs):
        server = cls(*args, **kwargs)
        while True:
            await asyncio.sleep(1.)

    @classmethod
    def createAndRun(cls, *args, **kwargs):
        from NaviNIBS.util.Asyncio import asyncioRunAndHandleExceptions
        asyncioRunAndHandleExceptions(cls.createAndRun_async, *args, **kwargs)


if __name__ == '__main__':
    ToolPositionsClient.createAndRun()



from __future__ import annotations

import shutil

import attrs
from datetime import datetime
import nibabel as nib
import json
import logging
import numpy as np
import os
import pandas as pd
import pyvista as pv
import tempfile
import typing as tp
from typing import ClassVar, TYPE_CHECKING

from NaviNIBS.util.attrs import attrsAsDict
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.numpy import array_equalish, attrsWithNumpyAsDict, attrsWithNumpyFromDict
if TYPE_CHECKING:
    from NaviNIBS.Navigator.Model.Session import Session

from NaviNIBS.Navigator.Model.GenericCollection import GenericCollection, GenericCollectionDictItem
from NaviNIBS.Navigator.Model.Calculations import calculateAngleFromMidlineFromCoilToMRITransf, calculateCoilToMRITransfFromTargetEntryAngle, getClosestPointToPointOnMesh


logger = logging.getLogger(__name__)


@attrs.define
class Target(GenericCollectionDictItem[str]):
    """
    Can specify (targetCoord, entryCoord, angle, [depthOffset]) to autogenerate coilToMRITransf,
    or (coilToMRITransf, [targetCoord]) to use transform directly.
    """
    _targetCoord: tp.Optional[np.ndarray] = None
    _entryCoord: tp.Optional[np.ndarray] = None
    _angle: tp.Optional[float] = None
    """
    Typical coil handle angle, defined relative to "midline"
    """
    _depthOffset: tp.Optional[float] = None
    """
    Offset beyond entryCoord, e.g. due to EEG electrode thickness, coil foam
    """
    _coilToMRITransf: tp.Optional[np.ndarray] = None
    """
    Uses convention where -y axis is along handle of typical coil and -z axis is pointing down into the head; origin is bottom face center of coil
    """

    _isVisible: bool = True
    """
    Whether currently visible in views that render subsets of targets
    """
    _isHistorical: bool = False
    """
    If True, will be hidden in almost all views, but may be referenced by samples (e.g. used when a target with associated samples is edited with a new orientation, but the associated samples should still be connected to the old version of the target)
    """
    _mayBeADependency: bool = False
    """
    Can be used to mark (e.g. when an associated sample is created) that this target is a dependency for other data elements, and a copy should be created and stored in history if this is edited
    """
    _isSelected: bool = False
    _color: str = '#0000FF'

    _cachedCoilToMRITransf: tp.Optional[np.ndarray] = attrs.field(init=False, default=None)
    """
    If coilToMRITransf isn't manually set, it might be auto-calculated from other target parameters.
    """
    _cachedCoilAngle: tp.Optional[float] = attrs.field(init=False, default=None)
    """
    If angle isn't manually set, it might be auto-calculated from other target parameters.
    """

    _session: tp.Optional[Session] = attrs.field(default=None, repr=False)
    """
    Used to access head model for determining angle from coilToMRITransf and vice versa
    """

    @property
    def targetCoord(self):
        return self._targetCoord

    @targetCoord.setter
    def targetCoord(self, newTargetCoord: tp.Optional[np.ndarray]):
        if array_equalish(self._targetCoord, newTargetCoord):
            return

        attribsChanging = ['targetCoord', 'coilToMRITransf']
        newCoilToMRITransf = calculateCoilToMRITransfFromTargetEntryAngle(
            session=self.session,
            targetCoord=newTargetCoord,
            entryCoord=self._entryCoord,
            angle=self.angle,  # note this may be autocalculated based on previous coilToMRITransf
            depthOffset=self._depthOffset
        )
        self.sigItemAboutToChange.emit(self.key, attribsChanging)
        self._targetCoord = newTargetCoord
        self._coilToMRITransf = None  # previous transform is invalid with new target
        self._cachedCoilToMRITransf = newCoilToMRITransf
        self.sigItemChanged.emit(self.key, attribsChanging)

    @property
    def entryCoord(self):
        return self._entryCoord

    @entryCoord.setter
    def entryCoord(self, newEntryCoord: tp.Optional[np.ndarray]):
        if array_equalish(self._entryCoord, newEntryCoord):
            return

        attribsChanging = ['entryCoord', 'coilToMRITransf']
        newCoilToMRITransf = calculateCoilToMRITransfFromTargetEntryAngle(
            session=self.session,
            targetCoord=self._targetCoord,
            entryCoord=newEntryCoord,
            angle=self.angle,  # note this may be autocalculated based on previous coilToMRITransf
            depthOffset=self._depthOffset
        )
        self.sigItemAboutToChange.emit(self.key, attribsChanging)
        self._entryCoord = newEntryCoord
        self._coilToMRITransf = None  # previous transform is invalid with new entry
        self._cachedCoilToMRITransf = newCoilToMRITransf
        self.sigItemChanged.emit(self.key, attribsChanging)

    @property
    def angle(self):
        """
        Angle from midline, in degrees.

        Can be manually specified. If not specified, will be autocalculated
        based on coilToMRITransf (if available)
        """
        if self._angle is not None:
            return self._angle
        else:
            return self.calculatedAngle

    @angle.setter
    def angle(self, newAngle: tp.Optional[float]):
        if self._angle == newAngle:
            return

        logger.debug(f'Changing angle from {self._angle} to {newAngle}')

        attribsChanging = ['angle', 'coilToMRITransf']
        newCoilToMRITransf = calculateCoilToMRITransfFromTargetEntryAngle(
            session=self.session,
            targetCoord=self._targetCoord,
            entryCoord=self._entryCoord,
            angle=newAngle,
            depthOffset=self._depthOffset
        )
        logger.debug(f'newCoilToMRITransf: {newCoilToMRITransf}')

        self.sigItemAboutToChange.emit(self.key, attribsChanging)
        self._angle = newAngle
        self._coilToMRITransf = None  # previous transform is invalid with new angle
        self._cachedCoilToMRITransf = newCoilToMRITransf
        self.sigItemChanged.emit(self.key, attribsChanging)

    @property
    def calculatedAngle(self) -> float:
        """
        Angle from midline (in degrees), calculated from coilToMRITransf (if available)
        """
        if self._cachedCoilAngle is None:
            if self._coilToMRITransf is not None:
                self._cachedCoilAngle = calculateAngleFromMidlineFromCoilToMRITransf(self.session, self._coilToMRITransf)
            elif self._cachedCoilToMRITransf is not None:
                self._cachedCoilAngle = calculateAngleFromMidlineFromCoilToMRITransf(self.session, self._cachedCoilToMRITransf)
            if self._cachedCoilAngle is None:
                if self._angle is not None:
                    return self._angle
                else:
                    return 0

        return self._cachedCoilAngle

    @property
    def depthOffset(self) -> float:
        return self._depthOffset if self._depthOffset is not None else 0.

    @depthOffset.setter
    def depthOffset(self, newDepthOffset: tp.Optional[float]):
        if self._depthOffset == newDepthOffset:
            return

        attribsChanging = ['depthOffset', 'coilToMRITransf']
        newCoilToMRITransf = calculateCoilToMRITransfFromTargetEntryAngle(
            session=self.session,
            targetCoord=self._targetCoord,
            entryCoord=self._entryCoord,
            angle=self.angle,  # note this may be autocalculated based on previous coilToMRITransf
            depthOffset=newDepthOffset
        )

        self.sigItemAboutToChange.emit(self.key, attribsChanging)
        self._depthOffset = newDepthOffset
        self._coilToMRITransf = None  # previous transform is invalid with new depth offset
        self._cachedCoilToMRITransf = newCoilToMRITransf
        self.sigItemChanged.emit(self.key, attribsChanging)

    @property
    def entryCoordPlusDepthOffset(self) -> tp.Optional[np.ndarray]:
        if self._entryCoord is None or self._targetCoord is None:
            if self._coilToMRITransf is not None:
                return self._coilToMRITransf[:3, 3].T
            return None
        if self._depthOffset is None or self._depthOffset == 0:
            return self._entryCoord

        entryVec = self._entryCoord - self._targetCoord
        entryVec /= np.linalg.norm(entryVec)

        offsetVec = entryVec * self._depthOffset

        return self._entryCoord + offsetVec

    @property
    def coilToMRITransf(self):
        if self._coilToMRITransf is not None:
            return self._coilToMRITransf
        else:
            if self._cachedCoilToMRITransf is None:
                self._cachedCoilToMRITransf = calculateCoilToMRITransfFromTargetEntryAngle(
                    session=self._session,
                    targetCoord=self.targetCoord,
                    entryCoord=self.entryCoord,
                    angle=self.angle,
                    depthOffset=self.depthOffset)
            return self._cachedCoilToMRITransf

    @coilToMRITransf.setter
    def coilToMRITransf(self, newCoilToMRITransf: tp.Optional[np.ndarray]):
        """
        Note that this may be set in a way that conflicts with targetCoord, entryCoord, angle, and depthOffset.
        However, if any of those are changed after setting this, this transf will be reset and be replaced
        with a cached auto-calculated version.
        """
        if array_equalish(self._coilToMRITransf, newCoilToMRITransf):
            return
        self.sigItemAboutToChange.emit(self.key, ['coilToMRITransf'])
        self._coilToMRITransf = newCoilToMRITransf
        self.sigItemChanged.emit(self.key, ['coilToMRITransf'])

    @property
    def isVisible(self):
        return self._isVisible

    @isVisible.setter
    def isVisible(self, isVisible: bool):
        if self._isVisible == isVisible:
            return
        self.sigItemAboutToChange.emit(self._key, ['isVisible'])
        self._isVisible = isVisible
        self.sigItemChanged.emit(self._key, ['isVisible'])

    @property
    def isHistorical(self):
        return self._isHistorical

    @isHistorical.setter
    def isHistorical(self, isHistorical: bool):
        if self._isHistorical == isHistorical:
            return
        self.sigItemAboutToChange.emit(self._key, ['isHistorical'])
        self._isHistorical = isHistorical
        self.sigItemChanged.emit(self._key, ['isHistorical'])

    @property
    def mayBeADependency(self):
        return self._mayBeADependency

    @mayBeADependency.setter
    def mayBeADependency(self, mayBeADependency: bool):
        if self._mayBeADependency == mayBeADependency:
            return
        self.sigItemAboutToChange.emit(self._key, ['mayBeADependency'])
        self._mayBeADependency = mayBeADependency
        self.sigItemChanged.emit(self._key, ['mayBeADependency'])

    @property
    def isSelected(self):
        return self._isSelected

    @isSelected.setter
    def isSelected(self, isSelected: bool):
        if self._isSelected == isSelected:
            return
        self.sigItemAboutToChange.emit(self.key, ['isSelected'])
        self._isSelected = isSelected
        self.sigItemChanged.emit(self.key, ['isSelected'])

    @property
    def color(self):
        return self._color

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, session: Session):
        if self._session is not session:
            self._session = session

    def autosetEntryCoord(self):
        """
        Note: this assumes the target coordinate is far inside the scalp (e.g. on cortical surface, not on the scalp itself)
        """
        closestPt_skin = getClosestPointToPointOnMesh(
            session=self._session,
            whichMesh='skinSurf',
            point_MRISpace=self.targetCoord)
        if closestPt_skin is None:
            raise ValueError('Missing information, cannot autoset entry coord')

        logger.info(f'Autosetting entry info to {closestPt_skin}')

        self.entryCoord = closestPt_skin

    def asDict(self) -> tp.Dict[str, tp.Any]:
        return attrsWithNumpyAsDict(self, npFields=('targetCoord', 'entryCoord', 'coilToMRITransf'), exclude=('session',))

    @classmethod
    def fromDict(cls, d: tp.Dict[str, tp.Any], session: Session | None = None):
        o = attrsWithNumpyFromDict(cls, d, npFields=('targetCoord', 'entryCoord', 'coilToMRITransf'))
        o.session = session

        if True:
            # auto-set entry coordinate based on target if entryCoord or coilToMRITransf not specified
            if o.targetCoord is not None and o.entryCoord is None and o._coilToMRITransf is None:
                if session is not None:
                    o.autosetEntryCoord()
                else:
                    logger.warning(f'No entry info for target {o.key}, no session info with which to autoset entry coord')

        return o


@attrs.define
class Targets(GenericCollection[str, Target]):
    _session: Session | None = attrs.field(default=None, repr=False)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.sigItemsChanged.connect(self._setSessionOnItemsChanged)

    def setWhichTargetsVisible(self, visibleKeys: tp.List[str]):
        self.setAttribForItems(self.keys(), dict(isVisible=[key in visibleKeys for key in self.keys()]))

    def setWhichTargetsSelected(self, selectedKeys: tp.List[str]):
        logger.debug(f'setWhichTargetsSelected: {selectedKeys}')
        self.setAttribForItems(self.keys(), dict(isSelected=[key in selectedKeys for key in self.keys()]))

    def _setSessionOnItemsChanged(self, keys: list[str], attribNames: list[str] | None = None):
        if attribNames is None:
            for key in keys:
                if key in self:
                    self[key].session = self._session

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, session: Session):
        if self._session is not session:
            self._session = session
            self._setSessionOnItemsChanged(self.keys())

    @classmethod
    def fromList(cls, itemList: list[dict[str, tp.Any]], session: Session | None = None) -> Targets:
        items = {}
        for itemDict in itemList:
            items[itemDict['key']] = Target.fromDict(itemDict, session=session)

        return cls(items=items, session=session)




from enum import Enum
import logging
from qtpy import QtWidgets, QtCore, QtGui

from NaviNIBS.util.Signaler import Signal


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class QFlowLayout(QtWidgets.QLayout):
    """
    A layout that arranges its children in a grid with a variable number of columns.
    """

    class LayoutMode(Enum):
        minimum = 0  # all items have minimum size
        equalMinimum = 1  # all items have equal size corresponding to minimum size of largest item
        equal = 2  # all items have equal size, filling available space

    def __init__(self, parent: QtWidgets.QWidget | None = None, margin: int = 0, spacing: int = -1, layoutMode: LayoutMode = LayoutMode.equal):
        super().__init__(parent)

        self.setSizeConstraint(QtWidgets.QLayout.SizeConstraint.SetMinimumSize)

        if parent is not None:
            self.setContentsMargins(margin, margin, margin, margin)

        self._itemList: list[QtWidgets.QLayoutItem] = []
        self._spacing = spacing
        self._layoutMode = layoutMode

        self._prevWidth: int | None = None
        self._prevHeight: int | None = None

        self.sigLayoutChanged: Signal[()] = Signal()

    def __del__(self):
        self._clear()

    def addItem(self, item: QtWidgets.QLayoutItem) -> None:
        self._itemList.append(item)

    def count(self) -> int:
        return len(self._itemList)

    def itemAt(self, index: int) -> QtWidgets.QLayoutItem | None:
        if 0 <= index < len(self._itemList):
            return self._itemList[index]
        return None

    def takeAt(self, index: int) -> QtWidgets.QLayoutItem | None:
        if 0 <= index < len(self._itemList):
            return self._itemList.pop(index)
        return None

    def hasHeightForWidth(self) -> bool:
        return True

    def heightForWidth(self, width: int) -> int:
        height = self._doLayout(QtCore.QRect(0, 0, width, 0), True)
        return height

    def expandingDirections(self):
        return QtCore.Qt.Orientations()

    def setGeometry(self, rect: QtCore.QRect) -> None:
        super().setGeometry(rect)
        height = self._doLayout(rect, False)
        if self._prevHeight != height:
            self._prevHeight = height
            self.update()
            # self.invalidate()
            self.sigLayoutChanged.emit()

    def sizeHint(self) -> QtCore.QSize:
        size = self.minimumSize()
        rect = self.geometry()
        if rect.width() > 0:
            layoutHeight = self._doLayout(rect, True)
            size.setHeight(layoutHeight)
        return size

    def minimumSize(self) -> QtCore.QSize:
        size = QtCore.QSize()

        for item in self._itemList:
            size = size.expandedTo(item.minimumSize())

        margins = self.contentsMargins()
        size += QtCore.QSize(margins.left() + margins.right(), margins.top() + margins.bottom())

        if True:
            width = self.contentsRect().width() + margins.left() + margins.right()
            if width <= 0 and self._prevWidth is not None:
                width = self._prevWidth
            size.setHeight(self.heightForWidth(width))

        return size

    def clear(self):
        self._clear()

    def _clear(self) -> None:
        while self.count():
            item = self.takeAt(0)

            if item is not None:
                if item.widget() is not None:
                    item.widget().deleteLater()

    def _doLayout(self, rect: QtCore.QRect, testOnly: bool) -> int:
        """
        Returns height
        """

        if not testOnly and rect.width() > 0:
            self._prevWidth = rect.width()

        x = rect.x() + self.contentsMargins().left()
        y = rect.y() + self.contentsMargins().top()
        lineHeight = 0

        match self._layoutMode:
            case QFlowLayout.LayoutMode.minimum:
                maxSpaceY = 0
                for item in self._itemList:
                    wid = item.widget()
                    spaceX = self.spacing() + wid.style().layoutSpacing(QtWidgets.QSizePolicy.PushButton,
                                                                          QtWidgets.QSizePolicy.PushButton,
                                                                         QtCore.Qt.Horizontal)
                    spaceY = self.spacing() + wid.style().layoutSpacing(QtWidgets.QSizePolicy.PushButton,
                                                                         QtWidgets.QSizePolicy.PushButton,
                                                                         QtCore.Qt.Vertical)
                    maxSpaceY = max(maxSpaceY, spaceY)

                    nextX = x + item.sizeHint().width() + spaceX

                    if nextX - spaceX > rect.right() - self.contentsMargins().right() and lineHeight > 0:
                        x = rect.x() + self.contentsRect().left()
                        y = y + lineHeight + spaceY
                        nextX = x + item.sizeHint().width()
                        lineHeight = 0

                    if not testOnly:
                        item.setGeometry(QtCore.QRect(QtCore.QPoint(x, y), item.sizeHint()))

                    x = nextX
                    lineHeight = max(lineHeight, item.sizeHint().height())

            case QFlowLayout.LayoutMode.equal | QFlowLayout.LayoutMode.equalMinimum:
                maxMinWidth = 0
                maxMinHeight = 0
                maxSpaceX = 0
                maxSpaceY = 0
                for item in self._itemList:
                    wid = item.widget()
                    maxMinWidth = max(maxMinWidth, item.sizeHint().width())
                    maxMinHeight = max(maxMinHeight, item.sizeHint().height())
                    maxSpaceX = max(maxSpaceX, self.spacing() + wid.style().layoutSpacing(QtWidgets.QSizePolicy.PushButton,
                                                                                          QtWidgets.QSizePolicy.PushButton,
                                                                                          QtCore.Qt.Horizontal))
                    maxSpaceY = max(maxSpaceY, self.spacing() + wid.style().layoutSpacing(QtWidgets.QSizePolicy.PushButton,
                                                                                          QtWidgets.QSizePolicy.PushButton,
                                                                                          QtCore.Qt.Vertical))

                if rect.width() == 0 and testOnly:
                    # special case: being tested without assigned size
                    return self.contentsMargins().top() + maxMinHeight + self.contentsMargins().bottom()

                try:
                    numCols = max((rect.width() - 1 - self.contentsMargins().left() - self.contentsMargins().right() - maxMinWidth), 0) // (maxMinWidth + maxSpaceX) + 1
                except ZeroDivisionError:
                    numCols = 1
                numRows = (len(self._itemList) - 1) // numCols + 1
                numEmptyLastRow = numCols * numRows - len(self._itemList)
                while numCols > 1 and numEmptyLastRow >= numRows:
                    numCols -= 1
                    numRows = (len(self._itemList) - 1) // numCols + 1
                    numEmptyLastRow = numCols * numRows - len(self._itemList)

                if not testOnly and self._layoutMode == QFlowLayout.LayoutMode.equalMinimum:
                    actualWidth = maxMinWidth
                else:
                    actualWidth = (rect.width() - 1 - maxSpaceX * (numCols - 1) - self.contentsMargins().left() - self.contentsMargins().right()) // numCols
                logger.debug(f'actualWidth: {actualWidth}')
                if actualWidth < 0:
                    logger.error('negative width')

                numInRow = 0
                numActualRows = 1
                for item in self._itemList:
                    wid = item.widget()
                    nextX = x + actualWidth + maxSpaceX

                    numInRow += 1

                    if numInRow > numCols or (nextX - maxSpaceX > rect.right() - self.contentsMargins().right() + 1 and lineHeight > 0):
                        x = rect.x() + self.contentsMargins().left()
                        y = y + lineHeight + maxSpaceY
                        nextX = x + actualWidth + maxSpaceX
                        lineHeight = 0
                        numInRow = 1
                        numActualRows += 1

                    if not testOnly:
                        item.setGeometry(QtCore.QRect(QtCore.QPoint(x, y), QtCore.QSize(actualWidth, maxMinHeight)))

                    x = nextX
                    lineHeight = max(lineHeight, maxMinHeight)

                logger.debug(f'numActualRows: {numActualRows}')

            case _:
                raise NotImplementedError('Invalid layout mode')

        logger.debug(f'lineHeight: {lineHeight}')

        if rect.width() < 0 and self._layoutMode != QFlowLayout.LayoutMode.equal:
            # calculated y is not meaningful
            return rect.y() + self.contentsMargins().top() + lineHeight + maxSpaceY + self.contentsMargins().bottom() - rect.y()
        else:
            height = y + lineHeight + self.contentsMargins().bottom() - rect.y()

        logger.debug(f'height: {height}')

        return height



if __name__ == '__main__':
    import NaviNIBS

    app = QtWidgets.QApplication([])
    win = QtWidgets.QMainWindow()
    win.setCentralWidget(QtWidgets.QWidget())
    superLayout = QtWidgets.QVBoxLayout()
    win.centralWidget().setLayout(superLayout)

    for layoutMode in (
            QFlowLayout.LayoutMode.minimum,
            QFlowLayout.LayoutMode.equal,
            QFlowLayout.LayoutMode.equalMinimum,
    ):
        w = QtWidgets.QGroupBox(f'Layout mode: {layoutMode.name}')
        layout = QFlowLayout(layoutMode=layoutMode)
        w.setLayout(layout)
        superLayout.addWidget(w)
        for i in range(13):
            layout.addWidget(QtWidgets.QPushButton(f'Button {i} {"-"*i}'))

        w.setSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Minimum)

    w = QtWidgets.QGroupBox('QGridLayout')
    layout = QtWidgets.QGridLayout()
    w.setLayout(layout)
    superLayout.addWidget(w)
    k = 0
    for i in range(4):
        for j in range(3):
            layout.addWidget(QtWidgets.QPushButton(f'Button {k} {"-"*k}'), i, j)
            k += 1

    superLayout.addStretch()

    win.show()

    win.adjustSize()

    app.exec_()



import asyncio
import attrs
import logging
import numpy as np
import qtawesome as qta
from qtpy import QtWidgets, QtCore, QtGui
import typing as tp

from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.Navigator.GUI.CollectionModels import CollectionTableModel, K, C, CI
from NaviNIBS.Navigator.GUI.CollectionModels.DigitizedLocationsTableModel import DigitizedLocationsTableModel
from NaviNIBS.Navigator.GUI.CollectionModels.FiducialsTableModels import PlanningFiducialsTableModel, RegistrationFiducialsTableModel
from NaviNIBS.Navigator.GUI.CollectionModels.HeadPointsTableModel import HeadPointsTableModel
from NaviNIBS.Navigator.GUI.CollectionModels.TargetsTableModel import TargetsTableModel, FullTargetsTableModel
from NaviNIBS.Navigator.GUI.CollectionModels.SamplesTableModel import SamplesTableModel
from NaviNIBS.Navigator.GUI.CollectionModels.ToolsTableModel import ToolsTableModel
from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.Navigator.Model.Samples import Sample, Samples
from NaviNIBS.Navigator.Model.Targets import Target, Targets
from NaviNIBS.Navigator.Model.Tools import Tool, Tools
from NaviNIBS.Navigator.Model.SubjectRegistration import HeadPoint, HeadPoints, Fiducial, Fiducials
from NaviNIBS.Navigator.Model.DigitizedLocations import DigitizedLocation, DigitizedLocations
from NaviNIBS.util.Signaler import Signal

logger = logging.getLogger(__name__)


TM = tp.TypeVar('TM', bound=CollectionTableModel)


@attrs.define
class CollectionTableWidget(tp.Generic[K, CI, C, TM]):
    _Model: tp.Callable[[Session], TM]
    _session: tp.Optional[Session] = attrs.field(default=None, repr=False)

    _tableView: QtWidgets.QTableView = attrs.field(init=False, factory=QtWidgets.QTableView)
    _model: tp.Optional[TM] = attrs.field(init=False, default=None)

    _doAdjustSizeToContents: bool = True

    _needsResizeToContents: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)
    _resizeToContentsPending: bool = attrs.field(init=False, default=False)

    sigCurrentItemChanged: Signal = attrs.field(init=False, factory=lambda: Signal((K,)))
    """
    Includes key (or index) of newly selected item.

    Note: this is emitted when the selection changes (e.g. a different sample is selected), NOT when a property of the currently selected sample changes.

    Note: this is Qt's "current" item. If multiple samples are selected, the "current" sample will just be the last sample added to the selection
    """

    sigSelectionChanged: Signal = attrs.field(init=False, factory=lambda: Signal((list[K],)))
    """
    Includes keys (or indices) of all selected items.

    Note: this is emitted when the selection changes (e.g. a different sample is selected), NOT when a property of the currently selected sample changes.
    """

    def __attrs_post_init__(self):
        self._tableView.setSelectionBehavior(self._tableView.SelectionBehavior.SelectRows)
        self._tableView.setSelectionMode(self._tableView.SelectionMode.ExtendedSelection)
        if self._doAdjustSizeToContents:
            self._tableView.setSizeAdjustPolicy(self._tableView.SizeAdjustPolicy.AdjustToContents)
            asyncio.create_task(asyncTryAndLogExceptionOnError(self._resizeToContentsLoop))

        """
        Allow drag reordering of columns. However, these won't affect underlying
        data models and won't be persistent.
        """
        self._tableView.horizontalHeader().setSectionsMovable(True)
        # TODO: add extra infrastructure to make these reorderings persistent
        # TODO: add right click context menu to hide / show columns

        if self._session is not None:
            self._onSessionSet()

    @property
    def wdgt(self):
        return self._tableView

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, newSes: tp.Optional[Session]):
        if self._session is newSes:
            return
        if self._session is not None:
            raise NotImplementedError  # TODO: notify table view of model change, disconnect from previous signals
        assert self._model is None
        self._session = newSes
        self._onSessionSet()

    def _onSessionSet(self):
        self._model = self._Model(session=self._session)
        self._model.sigSelectionChanged.connect(self._onModelSelectionChanged)
        self._tableView.setModel(self._model)
        self._tableView.selectionModel().currentChanged.connect(self._onTableCurrentChanged)
        self._tableView.selectionModel().selectionChanged.connect(self._onTableSelectionChanged)
        self._model.rowsInserted.connect(self._onTableRowsInserted)
        self._tableView.resizeColumnsToContents()

    @property
    def currentCollectionItemKey(self) -> tp.Optional[K]:
        curRow = self._tableView.currentIndex().row()
        if curRow == -1:
            return None
        if curRow >= self._model.rowCount():
            # can happen if rows were recently deleted
            return None
        return self._model.getCollectionItemKeyFromIndex(curRow)

    @currentCollectionItemKey.setter
    def currentCollectionItemKey(self, key: K | None):
        if key is None:
            if self.currentCollectionItemKey is not None:
                # current item was deleted, clear index
                self._tableView.setCurrentIndex(QtCore.QModelIndex())
            return
        if key == self.currentCollectionItemKey:
            return
        # logger.debug(f'Setting current item to {key}')
        index = self._model.getIndexFromCollectionItemKey(key)
        self._tableView.setCurrentIndex(self._model.index(index, 0))

    @property
    def rowCount(self):
        return self._model.rowCount()

    @property
    def currentRow(self):
        return self._tableView.currentIndex().row()

    @currentRow.setter
    def currentRow(self, row: int):
        if row == self.currentRow:
            return
        self._tableView.setCurrentIndex(self._model.index(row, 0))

    @property
    def currentCollectionItem(self) -> CI:
        curRow = self._tableView.currentIndex().row()
        return self._model.getCollectionItemFromIndex(curRow)

    @property
    def selectedCollectionItemKeys(self):
        selItems = self._tableView.selectedIndexes()
        selRows = [item.row() for item in selItems]
        selRows = list(dict.fromkeys(selRows))  # remove duplicates, keep order stable
        return [self._model.getCollectionItemKeyFromIndex(selRow) for selRow in selRows if selRow < self._model.rowCount()]

    def _onTableCurrentChanged(self):
        logger.debug('Current item changed')
        self.sigCurrentItemChanged.emit(self.currentCollectionItemKey)

    def _onTableSelectionChanged(self, selected: QtCore.QItemSelection, deselected: QtCore.QItemSelection):
        logger.debug('Current selection changed')
        selectedKeys = self.selectedCollectionItemKeys
        # logger.debug(f'selectedKeys: {selectedKeys}')
        if self.currentCollectionItemKey not in selectedKeys and len(selectedKeys) > 0:
            # change current item to first item in selected keys
            # logger.debug('Updating current item to be in selection')
            self.currentCollectionItemKey = selectedKeys[0]
        self._model.setWhichItemsSelected(selectedKeys)
        self.sigSelectionChanged.emit(selectedKeys)

    def _onTableRowsInserted(self, parent: QtCore.QModelIndex, first: int, last: int):
        # scroll to end of new rows automatically
        self._tableView.scrollTo(self._model.index(last, 0))
        self._needsResizeToContents.set()

    def _onModelSelectionChanged(self, changedKeys: list[K]):
        logger.debug(f'Updating selection for keys {changedKeys}')
        selection = QtCore.QItemSelection()
        selection.merge(self._tableView.selectionModel().selection(), QtCore.QItemSelectionModel.Select)

        for key in changedKeys:
            row = self._model.getIndexFromCollectionItemKey(key)
            if row is None:
                # item no longer in collection
                continue
            leftIndex = self._model.index(row, 0)
            rightIndex = self._model.index(row, self._model.columnCount() - 1)
            if self._model.getCollectionItemIsSelected(key):
                # logger.debug(f'{key} is selected')
                cmd = QtCore.QItemSelectionModel.Select
            else:
                # logger.debug(f'{key} is deselected')
                cmd = QtCore.QItemSelectionModel.Deselect
            selection.merge(QtCore.QItemSelection(leftIndex, rightIndex), cmd)
            # logger.debug(f'selection: {selection} {selection.indexes()}')

        self._tableView.selectionModel().select(selection, QtCore.QItemSelectionModel.ClearAndSelect)

    async def _resizeToContentsLoop(self):
        """
        resizeColumnsToContents is very expensive, so don't run on every update.
        Instead, wait until there are no changes for at least 20 sec to resize.
        """
        while True:
            await self._needsResizeToContents.wait()
            self._resizeToContentsPending = True
            while self._needsResizeToContents.is_set():
                self._needsResizeToContents.clear()
                await asyncio.sleep(20.)

            if not self._resizeToContentsPending:
                # someone manually resized while we were waiting
                continue

            logger.debug('Resizing columns to contents')
            self._tableView.resizeColumnsToContents()
            self._resizeToContentsPending = False

    def resizeColumnsToContents(self):
        """
        Allow caller to manually trigger resize without waititng for auto-resize loop.
        Note: this is an expensive operation
        """
        self._needsResizeToContents.clear()
        self._tableView.resizeColumnsToContents()
        self._resizeToContentsPending = False  # cancel any auto-queued resize


@attrs.define
class DigitizedLocationsTableWidget(CollectionTableWidget[str, DigitizedLocation, DigitizedLocations, DigitizedLocationsTableModel]):
    _Model: tp.Callable[[Session], DigitizedLocationsTableModel] = DigitizedLocationsTableModel

    def __attrs_post_init__(self):
        super().__attrs_post_init__()


@attrs.define
class SamplesTableWidget(CollectionTableWidget[str, Sample, Samples, SamplesTableModel]):
    _Model: tp.Callable[[Session], SamplesTableModel] = SamplesTableModel

    def __attrs_post_init__(self):
        super().__attrs_post_init__()


@attrs.define
class FullTargetsTableWidget(CollectionTableWidget[str, Target, Targets, FullTargetsTableModel]):
    _Model: tp.Callable[[Session], FullTargetsTableModel] = FullTargetsTableModel

    def __attrs_post_init__(self):
        super().__attrs_post_init__()


@attrs.define
class TargetsTableWidget(CollectionTableWidget[str, Target, Targets, TargetsTableModel]):
    _Model: tp.Callable[[Session], TargetsTableModel] = TargetsTableModel

    def __attrs_post_init__(self):
        super().__attrs_post_init__()


@attrs.define
class PlanningFiducialsTableWidget(CollectionTableWidget[str, Fiducial, Fiducials, PlanningFiducialsTableModel]):
    _Model: tp.Callable[[Session], PlanningFiducialsTableModel] = PlanningFiducialsTableModel

    def __attrs_post_init__(self):
        super().__attrs_post_init__()


@attrs.define
class RegistrationFiducialsTableWidget(CollectionTableWidget[str, Fiducial, Fiducials, RegistrationFiducialsTableModel]):
    _Model: tp.Callable[[Session], RegistrationFiducialsTableModel] = RegistrationFiducialsTableModel

    def __attrs_post_init__(self):
        super().__attrs_post_init__()


@attrs.define
class HeadPointsTableWidget(CollectionTableWidget[int, HeadPoint, HeadPoints, HeadPointsTableModel]):
    _Model: tp.Callable[[Session], HeadPointsTableModel] = HeadPointsTableModel

    def __attrs_post_init__(self):
        super().__attrs_post_init__()


@attrs.define
class ToolsTableWidget(CollectionTableWidget[int, Tool, Tools, ToolsTableModel]):
    _Model: tp.Callable[[Session], ToolsTableModel] = ToolsTableModel

    def __attrs_post_init__(self):
        super().__attrs_post_init__()


import asyncio
import attrs
import logging
import numpy as np
import typing as tp
from typing import ClassVar
import zmq
import zmq.asyncio as azmq

from NaviNIBS.Devices import positionsServerHostname, positionsServerPubPort, positionsServerCmdPort, TimestampedToolPosition
from NaviNIBS.util import ZMQAsyncioFix
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.logging import createLogFileHandler
from NaviNIBS.util.ZMQConnector import ZMQConnectorServer, logger as logger_ZMQConnector

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger_ZMQConnector.setLevel(logging.INFO)


@attrs.define
class ToolPositionsServer:
    """
    Base class for any tool positions server (to provide wrapper that is agnostic of connection type)
    """
    _type: ClassVar[str] = 'Generic'  # should be set by subclasses to be more informative

    _hostname: str = positionsServerHostname
    _pubPort: int = positionsServerPubPort
    _cmdPort: int = positionsServerCmdPort

    _logFilepath: str | None = None

    _latestPositions: tp.Dict[str, tp.Optional[TimestampedToolPosition]] = attrs.field(init=False, factory=dict)
    _publishingLatestLock: asyncio.Condition = attrs.field(init=False, factory=asyncio.Condition)
    _publishPending: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)
    _publishRateLimit: float = 10.  # in Hz

    _pubSocket: azmq.Socket = attrs.field(init=False)
    _connector: ZMQConnectorServer = attrs.field(init=False)
    _logFileHandler: logging.FileHandler = attrs.field(init=False)

    def __attrs_post_init__(self):
        ctx = azmq.Context()
        self._pubSocket = ctx.socket(zmq.PUB)
        logger.debug('Binding {}:{} for pub socket'.format(self._hostname, self._pubPort))
        self._pubSocket.bind('tcp://{}:{}'.format(self._hostname, self._pubPort))
        self._pubSocket.linger = 0  # TODO: determine if necessary

        self._connector = ZMQConnectorServer(
            obj=self,
            reqRepPort=self._cmdPort,
            bindAddr=self._hostname
        )

        if self._logFilepath is not None:
            self._logFileHandler = createLogFileHandler(self._logFilepath)
            logging.getLogger('').addHandler(self._logFileHandler)

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._publishLatestPositionsLoop))

    @property
    def type(self):
        return self._type

    async def run(self):
        raise NotImplementedError()  # should be implemented by subclass

    async def _publishLatestPositionsLoop(self):
        while True:
            await self._publishPending.wait()
            await asyncio.sleep(1/self._publishRateLimit)  # rate limit
            async with self._publishingLatestLock:
                logger.debug('Publishing latest positions')
                self._publishPending.clear()
                self._pubSocket.send_json({key: (val.asDict() if val is not None else None) for key, val in self._latestPositions.items()})

    async def recordNewPosition(self, key: str, position: TimestampedToolPosition | dict):
        if isinstance(position, dict):
            position = TimestampedToolPosition.fromDict(position)
        if key in self._latestPositions and self._latestPositions[key].time > position.time:
            logger.warning('New position appears to have been received out of order. Discarding.')
            return
        logger.debug('Received new position for {}: {}'.format(key, position))
        self._latestPositions[key] = position
        self._publishPending.set()

    def publishLatestPositions(self):
        self._publishPending.set()

    @classmethod
    async def createAndRun_async(cls, *args, **kwargs):
        server = cls(*args, **kwargs)
        await server.run()

    @classmethod
    def createAndRun(cls, *args, **kwargs):
        from NaviNIBS.util.Asyncio import asyncioRunAndHandleExceptions
        asyncioRunAndHandleExceptions(cls.createAndRun_async, *args, **kwargs)


import attrs
import numpy as np
import typing as tp

from NaviNIBS.util.attrs import attrsAsDict


def array_equalish(a: tp.Optional[np.ndarray], b: tp.Optional[np.ndarray], *args, **kwargs):
    """
    Similar to np.allclose but with shape comparison and None support like np.array_equal
    """

    if a is None or b is None:
        return a is None and b is None

    if not isinstance(a, np.ndarray) or not isinstance(b, np.ndarray):
        return a == b

    if not np.array_equal(a.shape, b.shape):
        return False

    # as a quick short-circuit test, compare the first element of array to avoid
    # running isclose on every array element if the first value is already different
    if a.size > 0 and not np.isclose(a.flat[0], b.flat[0], *args, **kwargs):
        return False

    return np.allclose(a, b, *args, **kwargs)


C = tp.TypeVar('C')


def attrsOptionalNDArrayField(init: bool = True) -> attrs.field:
    """
    Shorthand for an attrs field like ``x: np.ndarray | None = attrs.field(default=None)`` but with functional comparison behavior
    """
    return attrs.field(default=None,
                       init=init,
                       eq=attrs.cmp_using(
                           eq=array_equalish,
                           require_same_type=False))


def attrsWithNumpyAsDict(obj: C, npFields: tp.Optional[tp.Iterable[str]] = None,
                         eqs: tp.Optional[dict[str, tp.Callable]] = None,
                         **kwargs):
    """
    Helper function for serializing collections or collection items with numpy attributes.
    """
    if eqs is None:
        eqs = dict()
    if npFields is None:
        return attrsAsDict(obj, eqs=eqs, **kwargs)
    else:
        d = attrsAsDict(obj, eqs={field: array_equalish for field in npFields} | eqs, **kwargs)

        for key in npFields:
            if key in d and d[key] is not None:
                d[key] = d[key].tolist()

        return d


def attrsWithNumpyFromDict(cls: tp.Type[C], d: dict[str, tp.Any], npFields: tp.Optional[tp.Iterable[str]] = None, **kwargs) -> C:
    def convertOptionalNDArray(val: tp.Optional[tp.List[tp.Any]]) -> tp.Optional[np.ndarray]:
        if val is None:
            return None
        else:
            return np.asarray(val)

    for attrKey in npFields:
        if attrKey in d:
            d[attrKey] = convertOptionalNDArray(d[attrKey])

    return cls(**d, **kwargs)


import typing as tp
import attr
import contextlib

#  Note: once variadic generics are supported (see https://www.python.org/dev/peps/pep-0646/)
#   we'll be able to type hint this more usefully

ET = tp.TypeVarTuple('ET')
# Connection = tp.Callable[..., None]
Connection = tp.Callable[[*ET], None]


@attr.s(auto_attribs=True, eq=False)
class Signal(tp.Generic[*ET]):
    _types: tuple[*ET] = attr.ib(default=tuple())
    """
    Legacy, from before support for variadic generics was added.
    If documenting type with typing like Signal[T1, T2, T3], then don't specify a value for this attribute.
    """

    _connections: dict[int, set[Connection]] = attr.ib(init=False, factory=dict, repr=False)
    """
    Connections grouped by priority
    """

    _blockedSemaphoreCounter: int = attr.ib(init=False, default=0, repr=False)

    def __attrs_post_init__(self):
        pass

    def connect(self, fn: Connection, priority: int = 0):
        """
        Connections with higher priority are called first.
        Connections with same priority are called in undetermined order.
        """
        if priority not in self._connections:
            self._connections[priority] = set()

        self._connections[priority].add(fn)

    def disconnect(self, fn: Connection):
        for connectionSet in self._connections.values():
            connectionSet.remove(fn)

    @property
    def isBlocked(self):
        return self._blockedSemaphoreCounter > 0

    # def emit(self, *args, **kwargs) -> None:
    def emit(self, *args: *ET, **kwargs) -> None:
        if self._blockedSemaphoreCounter > 0:
            return
        priorities = sorted(self._connections.keys(), reverse=True)
        for priority in priorities:
            connectionSet = self._connections[priority].copy()
            for fn in connectionSet:
                if fn in self._connections[priority]:
                    fn(*args, **kwargs)

    @contextlib.contextmanager
    def blocked(self):
        """
        Can be used like:
                sig = Signal()
                sig.emit()  # will emit
                with sig.blocked():
                    sig.emit()  # will do nothing
                sig.emit()  # will emit
        """
        self._blockedSemaphoreCounter += 1
        try:
            yield None
        finally:
            self._blockedSemaphoreCounter -= 1

    @contextlib.contextmanager
    def connected(self, fn: Connection):
        self.connect(fn)
        try:
            yield None
        finally:
            self.disconnect(fn)

    @contextlib.contextmanager
    def disconnected(self, fn: Connection):
        assert any(fn in connectionSet for connectionSet in self._connections.values())
        self.disconnect(fn)
        try:
            yield None
        finally:
            self.connect(fn)



from __future__ import annotations

import logging
import numpy as np
import pyvista as pv
import pytransform3d.rotations as ptr
import pytransform3d.transformations as ptt
from skspatial.objects import Vector, Plane
import typing as tp
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.util.Transforms import applyTransform, composeTransform, invertTransform, estimateAligningTransform, concatenateTransforms, applyDirectionTransform
from NaviNIBS.util.pyvista.dataset import find_closest_point

logger = logging.getLogger(__name__)


def getClosestPointToPointOnMesh(session: Session, whichMesh: str, point_MRISpace: np.ndarray) -> tp.Optional[np.ndarray]:
    surf = getattr(session.headModel, whichMesh)
    if surf is None:
        return None

    assert isinstance(surf, pv.PolyData)

    # find closest point to coil on surf
    closestPtIndex = find_closest_point(surf, point_MRISpace)
    closestPt = surf.points[closestPtIndex, :]

    return closestPt


def calculateMidlineRefDirectionsFromCoilToMRITransf(session: Session, coilToMRITransf: np.ndarray | None) -> tuple[np.ndarray, np.ndarray] | tuple[None, None]:
    """
    Calculate the reference directions for angle=0 and angle=+90 degrees from midline, in the MRI space.

    Note that these directions are dependent on the coilToMRITransf, since the definition of "midline" can differ when on top of the head vs. extreme left/right vs. extreme anterior/posterior.

    :return
        refDir1: handle angle (i.e. coil's -y axis) corresponding to 0 degrees from midline
        refDir2: handle angle (i.e. coil's -y axis) corresponding to +90 degrees from midline
        May return (None, None) if coilToMRITransf is None.
    """

    if coilToMRITransf is None:
        return None, None

    # TODO: dynamically switch between MNI space and fiducial space depending on whether MNI transf is available
    if True:
        # use fiducial locations to define aligned coordinate space
        nas = session.subjectRegistration.fiducials.get('NAS', None)
        lpa = session.subjectRegistration.fiducials.get('LPA', None)
        rpa = session.subjectRegistration.fiducials.get('RPA', None)
        nas, lpa, rpa = tuple(fid.plannedCoord for fid in (nas, lpa, rpa))
        if any(coord is None for coord in (nas, lpa, rpa)):
            logger.debug('Missing fiducial(s), cannot find midline axis')
            return None, None

        centerPt = (lpa + rpa) / 2
        dirPA = nas - centerPt
        dirPA /= np.linalg.norm(dirPA)
        dirLR = rpa - lpa
        dirLR /= np.linalg.norm(dirLR)
        dirDU = np.cross(dirLR, dirPA)
        MRIToStdTransf = estimateAligningTransform(np.asarray([centerPt, centerPt + dirDU, centerPt + dirLR]),
                                                   np.asarray([[0, 0, 0], [0, 0, 1], [1, 0, 0]]))
    else:
        # TODO: use MNI transform to get midline points instead of assuming MRI is already aligned
        raise NotImplementedError

    coilLoc_stdSpace = applyTransform([coilToMRITransf, MRIToStdTransf], np.asarray([0, 0, 0]), doCheck=False)

    iDir = np.argmax(np.abs(coilLoc_stdSpace))
    match iDir:
        case 0:
            # far left or right
            refDir1 = np.asarray([0, -1, 0])  # this handle angle corresponds to 0 degrees from midline
            refDir2 = np.asarray([0, 0, -1]) * np.sign(coilLoc_stdSpace[iDir])  # this handle angle corresponds to +90 degrees from midline
        case 1:
            # far anterior or posterior
            refDir1 = np.asarray([0, 0, 1]) * np.sign(coilLoc_stdSpace[iDir])  # this handle angle corresponds to 0 degrees from midline
            refDir2 = np.asarray([1, 0, 0])  # this handle angle corresponds to +90 degrees from midline
        case 2:
            # far up (or down)
            refDir1 = np.asarray([0, -1, 0])  # this handle angle corresponds to 0 degrees from midline
            refDir2 = np.asarray([1, 0, 0]) * np.sign(coilLoc_stdSpace[iDir])  # this handle angle corresponds to +90 degrees from midline
        case _:
            raise NotImplementedError

    # convert refDirs to MRI space
    refDir1_MRI = applyDirectionTransform(invertTransform(MRIToStdTransf), refDir1, doCheck=False)
    refDir2_MRI = applyDirectionTransform(invertTransform(MRIToStdTransf), refDir2, doCheck=False)

    return refDir1_MRI, refDir2_MRI


def calculateAngleFromMidlineFromCoilToMRITransf(session: Session, coilToMRITransf: np.ndarray | None) -> float | None:
    if coilToMRITransf is None:
        return None

    refDir1_MRI, refDir2_MRI = calculateMidlineRefDirectionsFromCoilToMRITransf(session, coilToMRITransf)
    if refDir1_MRI is None or refDir2_MRI is None:
        return None

    handleDir_coilSpace = np.asarray([0, -1, 0])
    handleDir_MRI = applyDirectionTransform(coilToMRITransf, handleDir_coilSpace, doCheck=False)

    handleComp1 = np.dot(handleDir_MRI, refDir1_MRI)
    handleComp2 = np.dot(handleDir_MRI, refDir2_MRI)

    angle = np.arctan2(handleComp2, handleComp1)

    return np.rad2deg(angle).item()


def calculateCoilToMRITransfFromTargetEntryAngle(session: Session | None,
                                                 targetCoord: np.ndarray | None,
                                                 entryCoord: np.ndarray | None,
                                                 angle: float | None,
                                                 depthOffset: float | None,
                                                 prevCoilToMRITransf: np.ndarray | None = None)\
        -> np.ndarray | None:
    """
    Implemented outside of Target class to allow preparing for invalidating old coilToMRITransf when targetCoord, entryCoord, angle, or depthOffset change, before actually applying that change to the instance.
    """

    if session is None:
        return None

    if targetCoord is None:
        # estimate targetCoord from prevCoilToMRITransf
        coilOrigin_coilSpace = np.asarray([0, 0, 0])
        coilOrigin_MRISpace = applyTransform(prevCoilToMRITransf, coilOrigin_coilSpace, doCheck=False)
        pt_gm = getClosestPointToPointOnMesh(session=session,
                                             whichMesh='gmSurf',
                                             point_MRISpace=coilOrigin_MRISpace)
        # estimate gm pt by projecting same distance down coilToMRITransf depth axis
        coilToCortexDist = np.linalg.norm(coilOrigin_MRISpace - pt_gm)
        target_coilSpace = np.asarray([0, 0, -coilToCortexDist])
        targetCoord = applyTransform(prevCoilToMRITransf, target_coilSpace, doCheck=False)

    if entryCoord is None:
        coilOrigin_coilSpace = np.asarray([0, 0, 0])
        coilOrigin_MRISpace = applyTransform(prevCoilToMRITransf, coilOrigin_coilSpace, doCheck=False)
        entryCoord = coilOrigin_MRISpace

    if angle is None:
        angle = calculateAngleFromMidlineFromCoilToMRITransf(session=session, coilToMRITransf=prevCoilToMRITransf)

    logger.debug(f'Target angle: {angle}')

    if depthOffset is None:
        depthOffset = 0

    # determine rotation axis and angle to align with desired entry direction
    vec_targetDepth = Vector(entryCoord - targetCoord)
    vec_defaultCoilDepth = Vector([0, 0, 1])
    vec_rotAxis = vec_defaultCoilDepth.cross(vec_targetDepth)
    rotAngle = vec_targetDepth.angle_signed_3d(vec_defaultCoilDepth, vec_rotAxis)
    coilToMRITransf = np.eye(4)
    coilToMRITransf[:3, :3] = ptr.matrix_from_axis_angle(np.append(vec_rotAxis, -rotAngle))  # TODO: double check sign
    coilToMRITransf[:3, 3] = entryCoord
    coilToMRITransf[:3, 3] = applyTransform(coilToMRITransf, np.asarray([0, 0, depthOffset]), doCheck=False)

    # determine how much to rotate coil handle to get desired angle from midline
    refDir1_MRI, refDir2_MRI = calculateMidlineRefDirectionsFromCoilToMRITransf(session, coilToMRITransf)
    #logger.debug(f'refDir1_MRI: {refDir1_MRI} refDir2_MRI: {refDir2_MRI}')

    refDir1_coil = Vector(applyDirectionTransform(invertTransform(coilToMRITransf), refDir1_MRI, doCheck=False))
    refDir2_coil = Vector(applyDirectionTransform(invertTransform(coilToMRITransf), refDir2_MRI, doCheck=False))
    #logger.debug(f'refDir1_coil: {refDir1_coil} refDir2_coil: {refDir2_coil}')
    refPlaneNormal_coil = refDir1_coil.cross(refDir2_coil)
    #logger.debug(f'refPlaneNormal_coil: {refPlaneNormal_coil}')
    # note: this reference plane is likely tilted out of the XY plane of the coil space
    rotHandleInRefPlaneTransf = np.eye(4)
    rotHandleInRefPlaneTransf[:3, :3] = ptr.matrix_from_axis_angle(np.append(refPlaneNormal_coil, np.deg2rad(angle)))
    handleDirInRefPlane_coil = Vector(applyDirectionTransform(rotHandleInRefPlaneTransf, refDir1_coil, doCheck=False))
    #logger.debug(f'handleDirInRefPlane_coil: {handleDirInRefPlane_coil}')
    if True:
        vec_rotAxis = refPlaneNormal_coil.cross(handleDirInRefPlane_coil)
        projectWithinPlane = Plane.from_vectors(np.asarray([0, 0, 0]), handleDirInRefPlane_coil, refPlaneNormal_coil)
        xyPlane_coil = Plane.from_vectors(np.asarray([0, 0, 0]), np.asarray([1, 0, 0]), np.asarray([0, 1, 0]))
        handleLineInCoilPlane_coil = projectWithinPlane.intersect_plane(xyPlane_coil)
        handleDirInCoilPlane_coil = handleLineInCoilPlane_coil.direction
        if handleDirInCoilPlane_coil.dot(handleDirInRefPlane_coil) < 0:
            handleDirInCoilPlane_coil = -handleDirInCoilPlane_coil
    else:
        # project this handle direction onto the XY plane of the coil space
        handleDirInCoilPlane_coil = handleDirInRefPlane_coil.copy()
        handleDirInCoilPlane_coil[2] = 0
    logger.debug(f'handleDirInCoilPlane_coil: {handleDirInCoilPlane_coil}')

    initialHandleDir_coil = Vector([0, -1, 0])
    vec_rotAxis = np.asarray([0, 0, 1])  # TODO: check sign
    targetHandleAngle = initialHandleDir_coil.angle_signed_3d(handleDirInCoilPlane_coil, vec_rotAxis)
    logger.debug(f'targetHandleAngle: {np.rad2deg(targetHandleAngle)}')

    coilToMRITransf[:3, :3] = coilToMRITransf[:3, :3] @ ptr.matrix_from_axis_angle(np.append(vec_rotAxis, targetHandleAngle))

    if True:
        # TODO: debug, delete
        intermediateAngleFromMidline = calculateAngleFromMidlineFromCoilToMRITransf(session=session,
                                                                                    coilToMRITransf=coilToMRITransf)
        angleDiff = angle - intermediateAngleFromMidline
        logger.debug(f'angleFromMidline: {intermediateAngleFromMidline} angleDiff: {angleDiff}')
        if abs(angleDiff) > 10:
            logger.warning('Problem setting angle')

    logger.debug(f'newCoilToMRITransf: {coilToMRITransf}')

    return coilToMRITransf



from __future__ import annotations

import attrs
import logging
import numpy as np
import pyvista as pv
import typing as tp
from typing import ClassVar

from . import PlotViewLayer
from NaviNIBS.util.pyvista import setActorUserTransform

logger = logging.getLogger(__name__)


Transform = np.ndarray


@attrs.define(kw_only=True)
class TargetingAngleErrorLayer(PlotViewLayer):
    _type: ClassVar[str] = 'TargetingAngleError'

    _color: str = '#ff5500'
    _opacity: float = 0.5
    _lineWidth: float = 4.
    _radius: float = 10
    _multiplier: float = 1.

    _numArcSegments: int = 90  # TODO: check whether this needs to be reduced to improve render performance

    _angleMetric: str = 'Depth angle error'
    _angleOffset: float = -np.pi / 2
    _plotOnTargetOrCoil: str = 'target'
    _xyDims: tuple[int, int] = (0, 1)  # dimensions defining plane for angle visual

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._coordinator.sigCurrentTargetChanged.connect(lambda: self._queueRedraw(which='updateAngle'))
        self._coordinator.sigCurrentCoilPositionChanged.connect(lambda: self._queueRedraw(which=['updateAngle']))

    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str, ...]] = None):
        super()._redraw(which=which)

        if not isinstance(which, str):
            # assume parent call above triggered appropriate redraws
            return

        if which == 'all':
            which = ['initLine']
            self._redraw(which=which)
            return

        actorKey = self._getActorKey('line')

        if which == 'initLine':
            # start with dummy position, to be updated later
            pts_line = np.linspace([0, 0, 0], [1, 1, 1], self._numArcSegments + 1)

            self._actors[actorKey] = self._plotter.add_lines(pts_line,
                                                             connected=True,
                                                             color=self._color,
                                                             width=self._lineWidth,
                                                             name=actorKey)

            self._redraw(which='updateAngle')

        elif which == 'updateAngle':
            actor = self._actors[actorKey]

            hasTarget = self._coordinator.currentTarget
            hasCoil = self._coordinator.currentCoilToMRITransform is not None
            angle = self._coordinator.currentPoseMetrics.getValueForMetric(label=self._angleMetric)
            hasAngle = not np.isnan(angle)
            angle = np.deg2rad(angle) * self._multiplier

            doShow = hasTarget and hasCoil and hasAngle

            if doShow:
                pts_pv = self._getArcLines(startAngle=self._angleOffset,
                                           endAngle=angle + self._angleOffset,
                                           xyDims=self._xyDims,
                                           radius=self._radius,
                                           numPts=self._numArcSegments+1)
                with self._plotter.allowNonblockingCalls():
                    actor.GetMapper().SetInputData(pts_pv)

                    if self._plotOnTargetOrCoil == 'target':
                        currentTargetToMRITransform = self._coordinator.currentTarget.coilToMRITransf
                        setActorUserTransform(actor, currentTargetToMRITransform)

                    elif self._plotOnTargetOrCoil == 'coil':
                        currentCoilToMRITransform = self._coordinator.currentCoilToMRITransform
                        setActorUserTransform(actor, currentCoilToMRITransform)

                    else:
                        raise NotImplementedError

                if not actor.GetVisibility():
                    with self._plotter.allowNonblockingCalls():
                        actor.VisibilityOn()

            else:
                if actor.GetVisibility():
                    with self._plotter.allowNonblockingCalls():
                        actor.VisibilityOff()

            with self._plotter.allowNonblockingCalls():
                self._plotter.render()
        else:
            raise NotImplementedError(f'Unexpected redraw which: {which}')

    @classmethod
    def _getArcLines(cls, startAngle: float, endAngle: float,
                     xyDims: tuple[int, int],
                     radius: float, numPts: int = 300,) -> pv.PolyData:
        points = np.zeros((numPts, 3))
        theta = np.linspace(startAngle, endAngle, numPts)
        points[:, xyDims[0]] = radius * np.cos(theta)
        points[:, xyDims[1]] = radius * np.sin(theta)
        return pv.lines_from_points(points)